# Copyright (c) 2015-2016  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import binascii

"""Utility types used within geni-lib."""

class DPID(object):
  """Utility class representing OpenFlow Datapath IDs

  This class tries to handle all likely inputs and desired outputs, while
  providing a single internal type to work with in the code.

  String representations passed in must be represented in hex, but may contain
  common separators (colon, dash, and period) in any configuration.

  Args:
    val (int, long, unicode, str)

  Raises:
    DPID.OutOfRangeError: If the DPID represented by `val` is larger than the spec allows
      or less than zero.
    DPID.InputTypeError: If `val` is not a supported data type
  """

  MAX = (2 ** 64) - 1

  class OutOfRangeError(Exception):
    def __init__ (self, val):
      super(DPID.OutOfRangeError, self).__init__()
      self.val = val
    def __str__ (self):
      return "Input value (%d) out of range of valid DPIDs" % (self.val)

  class InputTypeError(Exception):
    def __init__ (self, val):
      super(DPID.InputTypeError, self).__init__()
      self.val = val
    def __str__ (self):
      return "Input value (%s) has invalid type (%s)" % (self.val, type(self.val))

  def __init__ (self, val):
    self._dpid = None

    if isinstance(val, (unicode, str)):
      val = int(val.translate(None, ":-."), 16)

    if isinstance(val, (int, long)):
      if val < DPID.MAX and val >= 0:
        self._dpid = val
      else:
        raise DPID.OutOfRangeError(val)
    else:
      raise DPID.InputTypeError(val)

  def __eq__ (self, other):
    return self._dpid == other._dpid # pylint: disable=W0212

  def __hash__ (self):
    return self._dpid

  def __str__ (self):
    """
    Returns:
      str: Hex formatted DPID, with colons
    """
    s = self.hexstr()
    return ":".join(["%s%s" % (s[x], s[x+1]) for x in xrange(0,15,2)])

  def __repr__ (self):
    return str(self)

  def __json__ (self):
    return str(self)

  def hexstr (self):
    """Unformatted hex representation of DPID

    Returns:
      str: Hex formatted DPID, without colons
    """
    return "%016x" % (self._dpid)

class EthernetMAC (object):
  """Utility class representing 48-bit Ethernet MAC Addresses

  This class tries to handle all likely inputs and desired outputs, while
  providing a single internal type to work with in the code.

  String representations passed in must be represented in hex, but may contain
  common separators (colon, dash, and period) in any configuration.

  Args:
    val (int, long, unicode, str)

  Raises:
    EthernetMAC.OutOfRangeError: If the MAC represented by `val` is larger than
      than 48-bits or less than zero.
    EthernetMAC.InputTypeError: If `val` is not a supported data type
  """

  MAX = 2 ** 48

  class OutOfRangeError(Exception):
    def __init__ (self, val):
      super(EthernetMAC.OutOfRangeError, self).__init__()
      self.val = val
    def __str__ (self):
      return "Input value (%d) out of range of valid Ethernet Addresses" % (self.val)

  class InputTypeError(Exception):
    def __init__ (self, val):
      super(EthernetMAC.InputTypeError, self).__init__()
      self.val = val
    def __str__ (self):
      return "Input value (%s) has invalid type (%s)" % (self.val, type(self.val))

  def __init__ (self, val):
    self._mac = None

    if isinstance(val, (unicode, str)):
      if len(val) == 6:
        val = binascii.hexlify(val)
      val = val.replace(":", "")
      val = val.replace("-", "")
      val = int(val, 16)

    if isinstance(val, (long, int)):
      if val < EthernetMAC.MAX and val >= 0:
        self._mac = val
      else:
        raise EthernetMAC.OutOfRangeError(val)
    else:
      raise EthernetMAC.InputTypeError(val)

  def __eq__ (self, other):
    return self._mac == other._mac # pylint: disable=W0212

  def __hash__ (self):
    return self._mac

  def __str__ (self):
    """
    Returns:
      str: Hex formatted MAC, with colons
    """
    s = self.hexstr()
    return ":".join(["%s%s" % (s[x], s[x+1]) for x in range(0,11,2)])

  def __json__ (self):
    return str(self)

  def __repr__ (self):
    return str(self)

  def hexstr (self):
    """Unformatted hex representation of MAC

    Returns:
      str: Hex formatted MAC, without separators
    """
    return "%012x" % (self._mac)



# Copyright (c) 2015-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# genish as an iPython extension for use with Jupyter

import copy
import itertools
import os.path
import types

import graphviz
import wrapt

from geni.aggregate.exceptions import AMError
from geni.aggregate.frameworks import KeyDecryptionError
from geni.aggregate.vts import VTS, HostPOAs, v4RouterPOAs
import geni.rspec.vtsmanifest
import geni.util
import geni.types

SHOW_ERROR_URL = False

######
### iPython-specific Utilities
######

def am_exc_handler (self, etype, value, tb, tb_offset = None):
  # The ipython docs for this handler are a lie - returning a structured traceback
  # is useless (ipython burns CPU cycles validating it, but never actually displays
  # any of it), you need to take care of all display yourself

  Colors = self.InteractiveTB.Colors


  out = []
  out.append("%s%s:%s %s" % (Colors.excName, etype.__name__, Colors.Normal, str(value)))

  if SHOW_ERROR_URL:
    try:
      out.append("AM Log URL: <%s>" % (value.error_url))
    except AttributeError:
      pass
  print "\n".join(out)


def topo (manifests, engine = "circo"):
  if not isinstance(manifests, list):
    manifests = [manifests]

  dotstr = geni.util.builddot(manifests)
  g = graphviz.Source(dotstr)
  g.engine = engine
  return g

LOGINROW = "<tr><td>{0}</td><td>{1}</td><td>{2}</td><td>{3}</td></tr>"
LOGINCOLS = ["Client-ID", "Username", "Host", "Port"]


def loginInfo (manifests):
  linfo = geni.util._corelogininfo(manifests)
  return RetListProxy(linfo, LOGINCOLS, LOGINROW, tupl = True)

def showErrorURL (show = False):
  global SHOW_ERROR_URL # pylint disable=global-statement
  SHOW_ERROR_URL = show


gsh = types.ModuleType("geni_ipython_util")
setattr(gsh, "showtopo", topo)
setattr(gsh, "printlogininfo", loginInfo)
setattr(gsh, "showErrorURL", showErrorURL)

#####
### Converters
#####

# Not actually supported right now, my brain doesn't want to figure it out
class Direction(object):
  TTB = 0
  LTR = 1

class ListGrid(object):
  def __init__ (self, iterable, cols, hdr, sort):
    self.iterable = iterable
    self.cols = cols
    self.header = ""
    self.sort = sort
    if hdr:
      self.header = """<tr><th colspan="%d" scope="row"><b>%s</b></th></tr>""" % (cols, hdr)
    rowc = ["<tr>"]
    for idx in xrange(cols):
      rowc.append("<td>%s</td>")
    rowc.append("</tr>")
    self.row = "".join(rowc)

  def _repr_html_ (self):
    if self.sort:
      contents = sorted([x for x in self.iterable])
      args = [iter(contents)] * self.cols
    else:
      args = [iter(self.iterable)] * self.cols
    rows = [self.row % tuple([str(y) for y in x]) for x in itertools.izip_longest(fillvalue="&nbsp;", *args)]
    out = """
    <table>
    %s
    %s
    </table>
    """ % (self.header, "\n".join(rows))
    return out

def listGridMaker (iterable, cols = 2, hdr = None, sort = False):
  return ListGrid(iterable, cols, hdr, sort)

setattr(gsh, "grid", listGridMaker)


STP_PORT_ROW = """<tr>
<td>%(client-id)s (%(num)d)</td><td>%(stp_state)s</td><td>%(stp_role)s</td><td>%(stp_port_id)s</td><td>%(stp_sec_in_state)s</td>
<td>%(stp_rx_count)d</td><td>%(stp_tx_count)d</td><td>%(stp_error_count)d</td>
</tr>"""

class STPProxy(wrapt.ObjectProxy):
  def _repr_html_ (self):
    brt = """
  <table>
    <tr><th colspan="3" scope="row">Bridge: <b>%(client-id)s</b></th></tr>
    <tr><th>Bridge ID</th><th>Designated Root</th><th>Root Path Cost</th></tr>
    <tr><td>%(stp_bridge_id)s</td><td>%(stp_designated_root)s</td><td>%(stp_root_path_cost)s</td></tr>
  </table>""" % (self)
    pelist = []
    for port in self["ports"]:
      d = {}
      d.update(port)
      d.update(port["info"])
      pe = STP_PORT_ROW % (d)
      pelist.append(pe)
    pt = """
  <table>
    <tr><th>Port</th><th>State</th><th>Role</th><th>Port ID</th><th>In State (secs)</th><th>RX</th><th>TX</th><th>Errors</th></tr>
    %s
  </table>
  """ % ("\n".join(pelist))
    return "%s\n%s" % (brt,pt)


RSTP_PORT_ROW = """<tr>
<td>%(client-id)s (%(num)d)</td><td>%(rstp_port_state)s</td><td>%(rstp_port_role)s</td><td>%(rstp_port_id)s</td>
<td>%(rstp_uptime)s</td><td>%(rstp_rx_count)d</td><td>%(rstp_tx_count)d</td><td>%(rstp_error_count)d</td>
</tr>"""

class RSTPProxy(wrapt.ObjectProxy):
  def _repr_html_ (self):
    brt = """
  <table>
    <tr><th colspan="3" scope="row">Bridge: <b>%(client-id)s</b></th></tr>
    <tr><th>Bridge ID</th><th>Designated Root</th><th>Root Path Cost</th></tr>
    <tr><td>%(rstp_bridge_id)s</td><td>%(rstp_root_id)s</td><td>%(rstp_root_path_cost)s</td></tr>
  </table>""" % (self)
    pelist = []
    for port in self["ports"]:
      d = {}
      d.update(port)
      d.update(port["info"])
      pe = RSTP_PORT_ROW % (d)
      pelist.append(pe)
    pt = """
  <table>
    <tr><th>Port</th><th>State</th><th>Role</th><th>Port ID</th><th>Uptime (secs)</th><th>RX</th><th>TX</th><th>Errors</th></tr>
    %s
  </table>
  """ % ("\n".join(pelist))
    return "%s\n%s" % (brt,pt)


def dictListBuilder (objlist, filter_cols, display_names):
  flist = []
  for obj in objlist:
    newl = []
    for col in filter_cols:
      newl.append(obj[col])
    flist.append(newl)

  rtemplate = "<tr>%s</tr>" % ("".join(["<td>{%d}</td>" % (x) for x in xrange(len(display_names))]))

  return RetListProxy(flist, display_names, rtemplate, True)


class RetListProxy(object):
  def __init__ (self, obj, columns, row_template, tupl = False):
    self._obj = obj
    self._columns = columns
    self._row_template = row_template
    self._col_template = "<th>%s</th>"
    self._tuple = tupl

  def __len__ (self):
    return len(self._obj)

  def __getitem__ (self, i):
    return self._obj[i]

  def __getslice__ (self, i, j):
    i = max(i, 0)
    j = max(j, 0)
    return self._obj[i:j]

  def __contains__ (self, item):
    return item in self._obj

  def __iter__ (self):
    for x in self._obj:
      yield x

  def _repr_html_ (self):
    trlist = []
    for row in self._obj:
      if self._tuple:
        trlist.append(self._row_template.format(*row))
      else:
        trlist.append(self._row_template.format(**row))

    collist = []
    for column in self._columns:
      collist.append(self._col_template % (column))

    return """<table>\n<tr>%s</tr>\n%s\n<table""" % ("".join(collist), "\n".join(trlist))
      

LEASEROW = "<tr><td>{hostname}</td><td>{ip-address}</td><td>{mac-address}</td><td>{binding-state}</td><td>{end:%Y-%m-%d %H:%M:%S}</td></tr>"
LEASECOLS = ["Hostname", "IP Address", "MAC Address", "State", "End"]

PINFOCOLS = ["Client ID", "ifindex", "vlan", "MTU", "Admin State", "Link State", "RX Bytes (Pkts)", "TX Bytes (Pkts)"]
PINFOROW = "<tr><td>{client-id}</td><td>{ifindex}</td><td>{tag}</td><td>{mtu}</td><td>{admin_state}</td><td>{link_state}</td><td>{statistics[rx_bytes]} ({statistics[rx_packets]})</td><td>{statistics[tx_bytes]} ({statistics[tx_packets]})</td></tr>"

FLOWCOLS = ["Table", "Duration", "Packets", "Bytes", "Rule"]
FLOWROW = "<tr><td>{table_id}</td><td>{duration}</td><td>{n_packets}</td><td>{n_bytes}</td><td>{rule}</td></tr>"

MACCOLS = ["Port", "VLAN", "MAC", "Age"]
MACROW = "<tr><td>{port}</td><td>{vlan}</td><td>{mac}</td><td>{age}</td></tr>"

#####
### Core geni-lib monkeypatches
#####

def replaceSymbol (module, name, func):
  """Moves module.name to module._name and sets module.name to the new function object."""
  setattr(module, "_%s" % (name), getattr(module, name))
  setattr(module, name, func)

def macTableDecomp (table):
  rowobjs = []
  for row in table[1:]:
    d = {}
    d["port"] = row[0]
    d["vlan"] = row[1]
    d["mac"] = geni.types.EthernetMAC(row[2])
    if row[3] is None:
      d["age"] = "-"
    else:
      d["age"] = int(row[3])
    rowobjs.append(d)
  return rowobjs

def flowTableDecomp (table):
  TEMPLATE = {"table_id" : 0, "duration" : None, "n_packets" : 0, "n_bytes" : None}
  rows = [[y.strip(",") for y in x.split(" ")] for x in table]
  rmaps = []
  for row in rows:
    rmap = copy.copy(TEMPLATE)
    for item in row[:-1]:
      (key,val) = item.split("=")
      rmap[key] = val
    rmap["rule"] = row[-1]
    rmaps.append(rmap)
  return rmaps

def dumpFlows (self, context, sname, datapaths, **kwargs):
  if not isinstance(datapaths, list):
    datapaths = [datapaths]

  res = self._dumpFlows(context, sname, datapaths, **kwargs)

  if len(res) == 1:
    return RetListProxy(flowTableDecomp(res.values()[0]), FLOWCOLS, FLOWROW)

  retd = {}
  for brname,table in res.items():
    retd[brname] = RetListProxy(flowTableDecomp(table), FLOWCOLS, FLOWROW)
  return retd

def getSTPInfo (self, context, sname, datapaths):
  if not isinstance(datapaths, list):
    datapaths = [datapaths]

  res = self._getSTPInfo(context, sname, datapaths)
  if len(res) == 1:
    return STPProxy(res[0])

  retobj = {}
  for br in res:
    retobj[br["client-id"]] = STPProxy(br)
  return retobj

def getRSTPInfo (self, context, sname, datapaths):
  if not isinstance(datapaths, list):
    datapaths = [datapaths]

  res = self._getRSTPInfo(context, sname, datapaths)
  if len(res) == 1:
    return RSTPProxy(res[0])

  retobj = {}
  for br in res:
    retobj[br["client-id"]] = RSTPProxy(br)
  return retobj


def getLeaseInfo (self, context, sname, client_ids):
  if not isinstance(client_ids, list):
    client_ids = [client_ids]

  res = self._getLeaseInfo(context, sname, client_ids)
  
  if len(res) == 1:
    return RetListProxy(res.values()[0], LEASECOLS, LEASEROW)
  
  retobj = {}
  for k,v in res.items():
    retobj[k] = RetListProxy(v, LEASECOLS, LEASEROW)
  return retobj

def getPortInfo (self, context, sname, client_ids):
  res = self._getPortInfo(context, sname, client_ids)

  if len(res) == 1:
    return RetListProxy(res.values()[0], PINFOCOLS, PINFOROW)

  retobj = {}
  for k,v in res.items():
    retobj[k] = RetListProxy(v, PINFOCOLS, PINFOROW)
  return retobj

def portDown (self, context, sname, port):
  if isinstance(port, (str, unicode)):
    pass
  elif isinstance(port, geni.rspec.vtsmanifest.GenericPort):
    port = port.client_id

  return self._portDown(context, sname, port)

def portUp (self, context, sname, port):                                                                                     
  if isinstance(port, (str, unicode)):                                                                                         
    pass
  elif isinstance(port, geni.rspec.vtsmanifest.GenericPort):                                                                   
    port = port.client_id                                                                                                      
              
  return self._portUp(context, sname, port)     

replaceSymbol(VTS, "dumpFlows", dumpFlows)
replaceSymbol(VTS, "getSTPInfo", getSTPInfo)
replaceSymbol(VTS, "getRSTPInfo", getRSTPInfo)
replaceSymbol(VTS, "getLeaseInfo", getLeaseInfo)
replaceSymbol(VTS, "getPortInfo", getPortInfo)
replaceSymbol(VTS, "portDown", portDown)
replaceSymbol(VTS, "portUp", portUp)


def getL2Table (self, context, sname, client_ids):
  if not isinstance(client_ids, list):
    client_ids = [client_ids]

  res = self._getL2Table(context, sname, client_ids)

  if len(res) == 1:
    return RetListProxy(macTableDecomp(res.values()[0]), MACCOLS, MACROW)

  retd = {}
  for l2d in res:
    for bridge, table in l2d.items():
      rowobjs = macTableDecomp(table)
      retd[bridge] = RetListProxy(rowobjs, MACCOLS, MACROW)

  return retd

def clearL2Table (self, context, sname, client_ids):
  if not isinstance(client_ids, list):
    client_ids = [client_ids]

  res = self._clearL2Table(context, sname, client_ids)

  return res

replaceSymbol(VTS, "getL2Table", getL2Table)
replaceSymbol(VTS, "clearL2Table", clearL2Table)


ARP_FILTER = ["hw-address", "ip-address", "status", "device"]
ARP_COLS = ["HW Address", "IP Address", "Status", "Interface"]
def getARPTable (self, context, sname, client_ids):
  res = self._getARPTable(context, sname, client_ids)

  if len(res.items()) == 1:
    return dictListBuilder(res.popitem()[1], ARP_FILTER, ARP_COLS)

  retobj = {}
  for k,v in res.items():
    retobj[k] = dictListBuilder(v, ARP_FILTER, ARP_COLS)
  return retobj
replaceSymbol(HostPOAs, "getARPTable", getARPTable)

ROUTE_FILTER = ["destination", "mask", "gateway", "device"]
ROUTE_COLS = ["Destination", "Mask", "Gateway", "Interface"]
def getRouteTable (self, context, sname, client_ids):
  res = self._getRouteTable(context, sname, client_ids)

  if len(res.items()) == 1:
    return dictListBuilder(res.popitem()[1], ROUTE_FILTER, ROUTE_COLS)

  retobj = {}
  for k,v in res.items():
    retobj[k] = dictListBuilder(v, ROUTE_FILTER, ROUTE_COLS)
  return retobj
replaceSymbol(HostPOAs, "getRouteTable", getRouteTable)

QROUTE_FILTER = ["type", "selected", "network", "next-hop", "interface", "time"]
QROUTE_COLS = ["", "Selected", "Network", "Next Hop", "Interface", "Duration"]
def getIPRouteTable (self, context, sname, client_ids):
  res = self._getRouteTable(context, sname, client_ids)
  if len(res.items()) == 1:
    return dictListBuilder(res.popitem()[1], QROUTE_FILTER, QROUTE_COLS)

  retobj = {}
  for k,v in res.items():
    retobj[k] = dictListBuilder(v, QROUTE_FILTER, QROUTE_COLS)
  return retobj
replaceSymbol(v4RouterPOAs, "getRouteTable", getIPRouteTable)

NEIGHBOR_FILTER = ["id", "priority", "state", "dead-time", "address", "interface"]
NEIGHBOR_COLS = ["ID", "Priority", "State", "Dead Time", "Address", "Interface"]
def getOSPFNeighbors (self, context, sname, client_ids):
  res = self._getOSPFNeighbors(context, sname, client_ids)
  if len(res.items()) == 1:
    return dictListBuilder(res.popitem()[1], NEIGHBOR_FILTER, NEIGHBOR_COLS)

  retobj = {}
  for k,v in res.items():
    retobj[k] = dictListBuilder(v, NEIGHBOR_FILTER, NEIGHBOR_COLS)
  return retobj
replaceSymbol(v4RouterPOAs, "getOSPFNeighbors", getOSPFNeighbors)

DNSRR_FILTER = ["type", "name", "value"]
DNSRR_COLS = ["Type", "Name", "Value"]
def getAllDNSResourceRecords(self, context, sname, client_ids):
  res = self._getAllDNSResourceRecords(context, sname, client_ids)
  if len(res.items()) == 1:
    return dictListBuilder(res.popitem()[1], DNSRR_FILTER, DNSRR_COLS)

  retobj = {}
  for k,v in res.items():
    retobj[k] = dictListBuilder(v, DNSRR_FILTER, DNSRR_COLS)
  return retobj
replaceSymbol(VTS, "getAllDNSResourceRecords", getAllDNSResourceRecords)

LASTDNSDHC_FILTER = ["data"]
LASTDNSDHC_COLS = ["Log"]
def getLastDNSDHCPops(self, context, sname, client_ids, count, type):
  res = self._getLastDNSDHCPops(context, sname, client_ids, count, type)
  if len(res.items()) == 1:
    return dictListBuilder(res.popitem()[1], LASTDNSDHC_FILTER, LASTDNSDHC_COLS)

  retobj = {}
  for k,v in res.items():
    retobj[k] = dictListBuilder(v, LASTDNSDHC_FILTER, LASTDNSDHC_COLS)
  return retobj
replaceSymbol(VTS, "getLastDNSDHCPops", getLastDNSDHCPops)

class NoContextError(Exception):
  def __str__ (self):
    return "No context found, extension cannot be loaded."

#####
### Extension loader
#####
def load_ipython_extension (ipy):
  import geni.util

  if not geni.util.hasDataContext():
    path = os.path.expanduser("~/omni.bundle")
    if os.path.exists(path):
      geni.util.buildContextFromBundle(path, pubkey_path = geni.util.MAKE_KEYPAIR)

  if not geni.util.hasDataContext():
    raise NoContextError()

  import geni._coreutil
  imports = geni._coreutil.shellImports()
  imports["genish"] = gsh

  import getpass

  tries = 0
  need_pw = False

  try:
    context = geni.util.loadContext(key_passphrase = "ffffffffff")
  except TypeError:
    context = geni.util.loadContext()
  except KeyDecryptionError:
    need_pw = True

  if need_pw:
    while True:
      pw = getpass.getpass("Private Key Passphrase: ")
      tries += 1
      try:
        imports["context"] = geni.util.loadContext(key_passphrase = pw)
        break
      except KeyDecryptionError:
        if tries < 3:
          continue
        break
      except IOError:
        break
  else:
    imports["context"] = context

  ipy.push(imports)
  ipy.set_custom_exc((AMError,), am_exc_handler)



# Copyright (c) 2013-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import inspect
import sys

from .protogeni import PGCompute

class IGCompute(PGCompute): pass

# TODO: Should warn if CMID from advertisement differs from one here

CaseWestern = IGCompute("ig-cwru", "boss.geni.case.edu", "urn:publicid:IDN+geni.case.edu+authority+cm")
CENIC = IGCompute("ig-cenic", "instageni.cenic.net", "urn:publicid:IDN+instageni.cenic.net+authority+cm")
Cornell = IGCompute("ig-cornell", "geni.it.cornell.edu", "urn:publicid:IDN+geni.it.cornell.edu+authority+cm")
Chicago = IGCompute("ig-chicago", "geni.uchicago.edu", "urn:publicid:IDN+geni.uchicago.edu+authority+cm")
Clemson = IGCompute("ig-clemson", "instageni.clemson.edu", "urn:publicid:IDN+instageni.clemson.edu+authority+cm")
Colorado = IGCompute("ig-colorado", "instageni.colorado.edu", "urn:publicid:IDN+instageni.colorado.edu+authority+cm")
Dublin = IGCompute("ig-ohmetrodc", "instageni.metrodatacenter.com", "urn:publicid:IDN+instageni.metrodatacenter.com+authority+cm")
GATech = IGCompute("ig-gatech", "instageni.rnoc.gatech.edu", "urn:publicid:IDN+instageni.rnoc.gatech.edu+authority+cm")
GPO = IGCompute("ig-gpo", "boss.instageni.gpolab.bbn.com", "urn:publicid:IDN+instageni.gpolab.bbn.com+authority+cm")
Illinois = IGCompute("ig-illinois", "instageni.illinois.edu", "urn:publicid:IDN+instageni.illinois.edu+authority+cm")
Kansas = IGCompute("ig-kansas", "instageni.ku.gpeni.net", "urn:publicid:IDN+instageni.ku.gpeni.net+authority+cm")
Kentucky = IGCompute("ig-kentucky", "boss.lan.sdn.uky.edu", "urn:publicid:IDN+lan.sdn.uky.edu+authority+cm")
Kettering = IGCompute("ig-kettering", "geni.kettering.edu", "urn:publicid:IDN+geni.kettering.edu+authority+cm")
LSU = IGCompute("ig-lsu", "instageni.lsu.edu", "urn:publicid:IDN+instageni.lsu.edu+authority+cm")
MAX = IGCompute("ig-max", "instageni.maxgigapop.net", "urn:publicid:IDN+instageni.maxgigapop.net+authority+cm")
Missouri = IGCompute("ig-missouri", "instageni.rnet.missouri.edu", "urn:publicid:IDN+instageni.rnet.missouri.edu+authority+cm")
MOXI = IGCompute("ig-moxi", "instageni.iu.edu", "urn:publicid:IDN+instageni.iu.edu+authority+cm")
Northwestern = IGCompute("ig-northwestern", "instageni.northwestern.edu",
                         "urn:publicid:IDN+instageni.northwestern.edu+authority+cm")
NPS = IGCompute("ig-nps", "instageni.nps.edu", "urn:publicid:IDN+instageni.nps.edu+authority+cm")
NYSERNet = IGCompute("ig-nysernet", "instageni.nysernet.org", "urn:publicid:IDN+instageni.nysernet.org+authority+cm")
NYU = IGCompute("ig-nyu", "genirack.nyu.edu", "urn:publicid:IDN+genirack.nyu.edu+authority+cm")
Princeton = IGCompute("ig-princeton", "instageni.cs.princeton.edu", "urn:publicid:IDN+instageni.cs.princeton.edu+authority+cm")
Rutgers = IGCompute("ig-rutgers", "instageni.rutgers.edu", "urn:publicid:IDN+instageni.rutgers.edu+authority+cm")
SOX = IGCompute("ig-sox", "instageni.sox.net", "urn:publicid:IDN+instageni.sox.net+authority+cm")
Stanford = IGCompute("ig-stanford", "instageni.stanford.edu", "urn:publicid:IDN+instageni.stanford.edu+authority+cm")
UCLA = IGCompute("ig-ucla", "instageni.idre.ucla.edu", "urn:publicid:IDN+instageni.idre.ucla.edu+authority+cm")
UKYMCV = IGCompute('ig-ukymcv', 'mcv.sdn.uky.edu', 'urn:publicid:IDN+mcv.sdn.uky.edu+authority+cm')
UKYPKS2 = IGCompute('ig-ukypks2', 'pks2.sdn.uky.edu', 'urn:publicid:IDN+pks2.sdn.uky.edu+authority+cm')
UMichigan = IGCompute('ig-umich', 'instageni.research.umich.edu', 'urn:publicid:IDN+instageni.research.umich.edu+authority+cm')
UMKC = IGCompute('ig-umkc', 'instageni.umkc.edu', 'urn:publicid:IDN+instageni.umkc.edu+authority+cm')
Utah = IGCompute("ig-utah", "boss.utah.geniracks.net", "urn:publicid:IDN+utah.geniracks.net+authority+cm")
UtahDDC = IGCompute("ig-utahddc", "boss.utahddc.geniracks.net", "urn:publicid:IDN+utahddc.geniracks.net+authority+cm")
UTC = IGCompute("ig-utc", "instageni.utc.edu", "urn:publicid:IDN+instageni.utc.edu+authority+cm")
UWashington = IGCompute("ig-uwashington", "instageni.washington.edu", "urn:publicid:IDN+instageni.washington.edu+authority+cm")
Wisconsin = IGCompute("ig-wisconsin", "instageni.wisc.edu", "urn:publicid:IDN+instageni.wisc.edu+authority+cm")
UKYMCV = IGCompute('ig-ukymcv', 'mcv.sdn.uky.edu', 'urn:publicid:IDN+mcv.sdn.uky.edu+authority+cm')

def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, PGCompute):
      yield obj

def name_to_aggregate ():
  result = dict()
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, PGCompute):
      result[obj.name] = obj
  return result

def cmid_to_aggregate ():
  result = dict()
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, PGCompute):
      result[obj._cmid] = obj
  return result



# Copyright (c) 2013-2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

class Image(object):
  def __init__ (self):
    self.id = None
    self.urn = None


class ImageRegistry(object):
  def __init__ (self):
    self._data = {}

  def register (self, manager, image):
    self._data.setdefault(image.id, {})[manager] = image



# Copyright (c) 2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from lxml import etree as ET

from .. import namespaces as GNS
from .pg import Namespaces as PGNS
from . import stitching

_XPNS = {'g' : GNS.REQUEST.name, 'e' : PGNS.EMULAB.name, 't' : stitching.STITCHNS.name}

class Advertisement(object):
  def __init__ (self, path = None, xml = None):
    if path:
      self._root = ET.parse(open(path))
    elif xml:
      self._root = ET.fromstring(xml)

  @property
  def stitchinfo (self):
    """Reference to the stitching info in the manifest, if present."""
    try:
      elem = self._root.xpath('/g:rspec/t:stitching', namespaces=_XPNS)[0]
      return stitching.AdInfo(elem)
    except IndexError:
      return None

  @property
  def text (self):
    return ET.tostring(self._root, pretty_print=True)



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from .pndefs import *
from .epcexc import *
from .epclan import *
from .epcnode import *
from .enodeb import *
from .ue import *
from .rflink import *



# Copyright (c) 2013-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import inspect
import sys

from .core import AM, APIRegistry

class PGCompute(AM):
  def __init__ (self, name, host, cmid = None, url = None):
    if url is None:
      url = "https://%s:12369/protogeni/xmlrpc/am/2.0" % (host)
    self.urlv3 = "%s3.0" % (url[:-3])
    self._apiv3 = APIRegistry.get("amapiv3")
    super(PGCompute, self).__init__(name, url, "amapiv2", "pg", cmid)

  def geniRestart (self, context, sname, urns):
    if not isinstance(urns, list):
      urns = [urns]
    return self._apiv3.poa(context, self.urlv3, sname, "geni_restart", urns)

  def geniStart (self, context, sname):
    return self._apiv3.poa(context, self.urlv3, sname, "geni_start")

  def geniUpdateUsers (self, context, sname, user_info_list):
    # user_info_list:
    # [ { "urn" : <str>, "keys" : [ <str>, ...] }, ... ]
    return self._apiv3.poa(context, self.urlv3, sname, "geni_update_users", options = {"geni_users" : user_info_list})

  def geniCancelUpdateUsers (self, context, sname):
    return self._apiv3.poa(context, self.urlv3, sname, "geni_updating_users_cancel")

  def getConsoleURL (self, context, sname, urn):
    return self._apiv3.poa(context, self.urlv3, sname, "geni_console_url", urns = [urn])


Kentucky_PG = PGCompute('pg-kentucky', 'www.uky.emulab.net', 'urn:publicid:IDN+uky.emulab.net+authority+cm')
UTAH_PG = PGCompute('pg-utah', 'www.emulab.net', 'urn:publicid:IDN+emulab.net+authority+cm')
Wall2_PG = PGCompute("pg-wall2", "www.wall2.ilabt.iminds.be", "urn:publicid:IDN+wall2.ilabt.iminds.be+authority+cm")
Wall1_PG = PGCompute("pg-wall1", "www.wall1.ilabt.iminds.be", "urn:publicid:IDN+wall1.ilabt.iminds.be+authority+cm")
wilab_PG = PGCompute("pg-wilab", "www.wilab2.ilabt.iminds.be", "urn:publicid:IDN+wilab2.ilabt.iminds.be+authority+cm")

def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      yield obj

def name_to_aggregate ():
  result = dict()
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      result[obj.name] = obj
  return result



# Copyright (c) 2013-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import itertools
import sys
import functools
import importlib

from lxml import etree as ET

import geni.rspec
import geni.namespaces as GNS
import geni.urn

# This exception gets thrown if a __ONCEONLY__ extension gets added to a parent
# element more than one time
class DuplicateExtensionError(Exception):
  def __init__ (self, klass):
    super(DuplicateExtensionError, self).__init__()
    self.klass = klass
  def __str__ (self):
    return "Extension (%s) can only be added to a parent object once" % self.klass.__name__

################################################
# Base Request - Must be at top for EXTENSIONS #
################################################

class Request(geni.rspec.RSpec):
  EXTENSIONS = []

  def __init__ (self):
    super(Request, self).__init__("request")
    self._resources = []
    self.tour = None
    self._raw_elements = []

    self.addNamespace(GNS.REQUEST, None)
    self.addNamespace(Namespaces.CLIENT)

    self._ext_children = []
    for name,ext in Request.EXTENSIONS:
      self._wrapext(name,ext)

  def _wrapext (self, name, klass):
    @functools.wraps(klass.__init__)
    def wrap(*args, **kw):
      if getattr(klass, "__ONCEONLY__", False):
        if any(map(lambda x: isinstance(x,klass),self._ext_children)):
          raise DuplicateExtensionError(klass)
      instance = klass(*args, **kw)
      if getattr(klass, "__WANTPARENT__", False):
        instance._parent = self
      self._ext_children.append(instance)
      return instance
    setattr(self, name, wrap)

  def addResource (self, rsrc):
    for ns in rsrc.namespaces:
      self.addNamespace(ns)
    self._resources.append(rsrc)

  @property
  def resources(self):
    return self._resources + self._ext_children

  def addTour (self, tour):
    self.addNamespace(Namespaces.EMULAB)
    self.addNamespace(Namespaces.JACKS)
    self.tour = tour

  def hasTour (self):
    return self.tour is not None

  def addRawElement (self, elem):
    self._raw_elements.append(elem)

  def writeXML (self, path):
    """Write the current request contents as an XML file that represents an rspec
    in the GENIv3 format."""

    if path is None:
      f = sys.stdout
    else:
      f = open(path, "w+")

    buf = self.toXMLString(True)
    f.write(buf)

    if path is not None:
      f.close()

  def toXMLString (self, pretty_print = False):
    """Return the current request contents as an XML string that represents an rspec
    in the GENIv3 format."""

    rspec = self.getDOM()

    if self.tour:
      self.tour._write(rspec)

    for resource in self._resources:
      resource._write(rspec)

    for obj in self._ext_children:
      obj._write(rspec)

    for elem in self._raw_elements:
      rspec.append(elem)

    buf = ET.tostring(rspec, pretty_print = pretty_print)
    return buf



class Resource(object):
  def __init__ (self):
    self.namespaces = []
    self._ext_children = []

  def addNamespace (self, ns):
    self.namespaces.append(ns)

  def _wrapext (self, name, klass):
    @functools.wraps(klass.__init__)
    def wrap(*args, **kw):
      if getattr(klass, "__ONCEONLY__", False):
        if any(map(lambda x: isinstance(x,klass),self._ext_children)):
          raise DuplicateExtensionError(klass)
      instance = klass(*args, **kw)
      self._ext_children.append(instance)
      return instance
    setattr(self, name, wrap)

  def _write (self, element):
    for obj in self._ext_children:
      obj._write(element)
    return element


class NodeType(object):
  XEN = "emulab-xen"
  RAW = "raw"
  VM = "emulab-xen"


class Command(object):
  def __init__ (self, cmd, data):
    self.cmd = cmd
    self.data = data

  def resolve (self):
    return self.cmd % self.data


class Service(object):
  def __init__ (self):
    pass


class Install(Service):
  def __init__ (self, url, path):
    super(Install, self).__init__()
    self.url = url
    self.path = path

  def _write (self, element):
    ins = ET.SubElement(element, "{%s}install" % (GNS.REQUEST.name))
    ins.attrib["url"] = self.url
    ins.attrib["install_path"] = self.path
    return ins


class Execute(Service):
  def __init__ (self, shell, command):
    super(Execute, self).__init__()
    self.shell = shell
    self.command = command

  def _write (self, element):
    exc = ET.SubElement(element, "{%s}execute" % (GNS.REQUEST.name))
    exc.attrib["shell"] = self.shell
    if isinstance(self.command, Command):
      exc.attrib["command"] = self.command.resolve()
    else:
      exc.attrib["command"] = self.command
    return exc


class Address(object):
  def __init__ (self, atype):
    self.type = atype


class IPv4Address(Address):
  def __init__ (self, address, netmask):
    super(IPv4Address, self).__init__("ipv4")
    self.address = address
    self.netmask = netmask

  def _write (self, element):
    ip = ET.SubElement(element, "{%s}ip" % (GNS.REQUEST.name))
    ip.attrib["address"] = self.address
    ip.attrib["netmask"] = self.netmask
    ip.attrib["type"] = self.type
    return ip


class Interface(object):
  class InvalidAddressTypeError(Exception):
    def __init__ (self, addr):
      super(Interface.InvalidAddressTypeError, self).__init__()
      self.addr = addr
    def __str__ (self):
      return "Type (%s) is invalid for interface addresses." % (type(self.addr))

  def __init__ (self, name, node, address = None):
    self.client_id = name
    self.node = node
    self.addresses = []
    self.component_id = None
    self.bandwidth = None
    self.latency = None
    self.plr = None
    if address:
      self.addAddress(address)

  @property
  def name (self):
    return self.client_id

  def addAddress (self, address):
    if isinstance(address, Address):
      self.addresses.append(address)
    else:
      raise Interface.InvalidAddressTypeError(address)

  def _write (self, element):
    intf = ET.SubElement(element, "{%s}interface" % (GNS.REQUEST.name))
    intf.attrib["client_id"] = self.client_id
    if self.component_id:
      if isinstance(self.component_id, geni.urn.Base):
        intf.attrib["component_id"] = str(self.component_id)
      else:
        intf.attrib["component_id"] = self.component_id
    for addr in self.addresses:
      addr._write(intf)
    return intf


class Link(Resource):
  EXTENSIONS = []
  LNKID = 0
  DEFAULT_BW = -1
  DEFAULT_LAT = 0
  DEFAULT_PLR = 0.0

  def __init__ (self, name = None, ltype = "", members = None):
    super(Link, self).__init__()
    if name is None:
      self.client_id = Link.newLinkID()
    else:
      self.client_id = name

    self.interfaces = []

    if members is not None:
      for member in members:
        if isinstance(member,Interface):
          self.addInterface(member)
        else:
          self.addInterface(member.addInterface())

    self.type = ltype
    self.shared_vlan = None
    self._mac_learning = True
    self._vlan_tagging = None
    self._trivial_ok = None
    self._link_multiplexing = False
    self._best_effort = False
    self._ext_children = []
    self._raw_elements = []
    self._component_managers = []
    self.protocol = None

    # If you try to set bandwidth higher than a gigabit, PG probably won't like you
    self.bandwidth = Link.DEFAULT_BW
    self.latency = Link.DEFAULT_LAT
    self.plr = Link.DEFAULT_PLR

    for name,ext in Link.EXTENSIONS:
      self._wrapext(name,ext)

  def _wrapext (self, name, klass):
    @functools.wraps(klass.__init__)
    def wrap(*args, **kw):
      if getattr(klass, "__ONCEONLY__", False):
        if any(map(lambda x: isinstance(x,klass),self._ext_children)):
          raise DuplicateExtensionError(klass)
      instance = klass(*args, **kw)
      if getattr(klass, "__WANTPARENT__", False):
        instance._parent = self
      self._ext_children.append(instance)
      return instance
    setattr(self, name, wrap)

  def addRawElement (self, elem):
    self._raw_elements.append(elem)

  @classmethod
  def newLinkID (cls):
    Link.LNKID += 1
    return "link-%d" % (Link.LNKID)

  def addChild (self, obj):
    self._ext_children.append(obj)

  def addInterface (self, intf):
    self.interfaces.append(intf)

  def addNode (self, node):
    interface = node.addInterface()
    self.interfaces.append(interface)
    return interface

  def addComponentManager (self, component_manager):
    self._component_managers.append(component_manager)

  def connectSharedVlan (self, name):
    self.namespaces.append(GNS.SVLAN)
    self.shared_vlan = name

  def disableMACLearning (self):
    self.namespaces.append(Namespaces.VTOP)
    self._mac_learning = False

  def enableVlanTagging (self):
    import geni.warnings as GW
    import warnings
    warnings.warn("Link.enableVlanTagging() is deprecated, please use the Link.vlan_tagging attribute instead.",
                  GW.GENILibDeprecationWarning, 2)
    self.vlan_tagging = True

  @property
  def vlan_tagging (self):
    return self._vlan_tagging

  @vlan_tagging.setter
  def vlan_tagging (self, val):
    self.namespaces.append(Namespaces.EMULAB)
    self._vlan_tagging = val

  @property
  def best_effort (self):
    return self._best_effort

  @best_effort.setter
  def best_effort (self, val):
    self.namespaces.append(Namespaces.EMULAB)
    self._best_effort = val

  @property
  def link_multiplexing (self):
    return self._link_multiplexing

  @link_multiplexing.setter
  def link_multiplexing (self, val):
    self.namespaces.append(Namespaces.EMULAB)
    self._link_multiplexing = val

  @property
  def trivial_ok (self):
    return self._trivial_ok

  @trivial_ok.setter
  def trivial_ok (self, val):
    self.namespaces.append(Namespaces.EMULAB)
    self._trivial_ok = val

  def _write (self, root):
    # pylint: disable=too-many-branches
    lnk = ET.SubElement(root, "{%s}link" % (GNS.REQUEST.name))
    lnk.attrib["client_id"] = self.client_id
    if self.protocol:
      lnk.attrib["protocol"] = self.protocol

    for intf in self.interfaces:
      ir = ET.SubElement(lnk, "{%s}interface_ref" % (GNS.REQUEST.name))
      ir.attrib["client_id"] = intf.client_id
    if self.type != "":
      lt = ET.SubElement(lnk, "{%s}link_type" % (GNS.REQUEST.name))
      lt.attrib["name"] = self.type
    if self.shared_vlan:
      sv = ET.SubElement(lnk, "{%s}link_shared_vlan" % (GNS.SVLAN.name))
      sv.attrib["name"] = self.shared_vlan

    if not self._mac_learning:
      lrnelem = ET.SubElement(lnk, "{%s}link_attribute" % (Namespaces.VTOP.name))
      lrnelem.attrib["key"] = "nomac_learning"
      lrnelem.attrib["value"] = "yep"

    # Do not want to spit out a vlan tagging statement unless explicitly
    # set, since the default is no tagging and always spitting that out
    # will be bad for regression testing.
    if self._vlan_tagging != None:
      tagging = ET.SubElement(lnk, "{%s}vlan_tagging" % (Namespaces.EMULAB.name))
      if self._vlan_tagging:
        tagging.attrib["enabled"] = "true"
      else:
        tagging.attrib["enabled"] = "false"

    if self._best_effort:
      tagging = ET.SubElement(lnk, "{%s}best_effort" % (Namespaces.EMULAB.name))
      tagging.attrib["enabled"] = "true"

    if self._link_multiplexing:
      tagging = ET.SubElement(lnk, "{%s}link_multiplexing" % (Namespaces.EMULAB.name))
      tagging.attrib["enabled"] = "true"

    if self._trivial_ok is not None:
      trivial = ET.SubElement(lnk, "{%s}trivial_ok" % (Namespaces.EMULAB.name))
      if self._trivial_ok:
        trivial.attrib["enabled"] = "true"
      else:
        trivial.attrib["enabled"] = "false"

    # LAN shaping properties are handled by the LAN class below.
    if self.type != "lan":
      for (intfA,intfB) in itertools.permutations(self.interfaces,2):
        bw = intfA.bandwidth if intfA.bandwidth else self.bandwidth
        lat = intfA.latency if intfA.latency else self.latency
        plr = intfA.plr if intfA.plr else self.plr
        self._write_link_prop(lnk, intfB.client_id, intfA.client_id, bw, lat, plr)

    for obj in self._ext_children:
      obj._write(lnk)

    for elem in self._raw_elements:
      lnk.append(elem)

    for manager in self._component_managers:
      cm = ET.SubElement(lnk, "{%s}component_manager" % (GNS.REQUEST.name))
      cm.attrib["name"] = manager

    return lnk

  def _write_link_prop(self, lnk, src, dst, bw, lat, plr):
    if bw != Link.DEFAULT_BW or lat != Link.DEFAULT_LAT or \
       plr != Link.DEFAULT_PLR:
      prop = ET.SubElement(lnk, "{%s}property" % (GNS.REQUEST.name))
      prop.attrib["source_id"] = src
      prop.attrib["dest_id"] = dst
      if bw != Link.DEFAULT_BW:
        prop.attrib["capacity"] = str(bw)
      if lat != Link.DEFAULT_LAT:
        prop.attrib["latency"] = str(lat)
      if plr != Link.DEFAULT_PLR:
        prop.attrib["packet_loss"] = str(plr)

Request.EXTENSIONS.append(("Link", Link))


class LAN(Link):
  def __init__ (self, name = None):
    super(LAN, self).__init__(name, "lan")

  def _write (self, root):
    lnk = super(LAN, self)._write(root)

    for intf in self.interfaces:
      bw = intf.bandwidth if intf.bandwidth else self.bandwidth
      lat = intf.latency if intf.latency else self.latency
      plr = intf.plr if intf.plr else self.plr
      super(LAN, self)._write_link_prop(lnk, intf.client_id, self.client_id, bw, lat, plr)

    return lnk

Request.EXTENSIONS.append(("LAN", LAN))


class L3GRE(Link):
  def __init__ (self, name = None):
    super(L3GRE, self).__init__(name, "gre-tunnel")

Request.EXTENSIONS.append(("L3GRE", L3GRE))

class L2GRE(Link):
  def __init__ (self, name = None):
    super(L2GRE, self).__init__(name, "egre-tunnel")

Request.EXTENSIONS.append(("L2GRE", L2GRE))

class StitchedLink(Link):
  class UnknownComponentManagerError(Exception):
    def __init__ (self, cid):
      super(StitchedLink.UnknownComponentManagerError, self).__init__()
      self._cid = cid
    def __str__ (self):
      return "Interface with client_id %s is not attached to a bound node." % (self._cid)

  class TooManyInterfacesError(Exception):
    def __str__ (self):
      return "Stitched Links may not be connected to more than two interfaces"

  def __init__ (self, name = None):
    super(StitchedLink, self).__init__(name, "")
    self.bandwidth = 20000

  def _write (self, root):
    if len(self.interfaces) > 2:
      raise StitchedLink.TooManyInterfacesError()

    lnk = super(StitchedLink, self)._write(root)
    for intf in self.interfaces:
      if intf.node.component_manager_id is None:
        raise StitchedLink.UnknownComponentManagerError(intf.client_id)
      cm = ET.SubElement(lnk, "{%s}component_manager" % (GNS.REQUEST.name))
      cm.attrib["name"] = intf.node.component_manager_id
    return lnk

Request.EXTENSIONS.append(("StitchedLink", StitchedLink))

class Node(Resource):
  EXTENSIONS = []

  def __init__ (self, name, ntype, component_id = None, exclusive = None):
    super(Node, self).__init__()
    self.client_id = name
    self.exclusive = exclusive
    self.disk_image = None
    self.type = ntype
    self.hardware_type = None
    self.interfaces = []
    self.services = []
    self.routable_control_ip = False
    self.component_id = component_id
    self.component_manager_id = None
    self._ext_children = []
    for name,ext in Node.EXTENSIONS:
      self._wrapext(name,ext)

    self._raw_elements = []

  class DuplicateInterfaceName(Exception):
    def __str__ (self):
      return "Duplicate interface names"

  def _wrapext (self, name, klass):
    @functools.wraps(klass.__init__)
    def wrap(*args, **kw):
      if getattr(klass, "__ONCEONLY__", False):
        if any(map(lambda x: isinstance(x,klass),self._ext_children)):
          raise DuplicateExtensionError(klass)
      instance = klass(*args, **kw)
      if getattr(klass, "__WANTPARENT__", False):
        instance._parent = self
      self._ext_children.append(instance)
      return instance
    setattr(self, name, wrap)

  @property
  def name (self):
    return self.client_id

  def _write (self, root):
    # pylint: disable=too-many-branches
    nd = ET.SubElement(root, "{%s}node" % (GNS.REQUEST.name))
    nd.attrib["client_id"] = self.client_id
    if self.exclusive is not None:  # Don't write this for EG
      nd.attrib["exclusive"] = str(self.exclusive).lower()
    if self.component_id:
      if isinstance(self.component_id, geni.urn.Base):
        nd.attrib["component_id"] = str(self.component_id)
      else:
        nd.attrib["component_id"] = self.component_id
    if self.component_manager_id:
      if isinstance(self.component_manager_id, geni.urn.Base):
        nd.attrib["component_manager_id"] = str(self.component_manager_id)
      else:
        nd.attrib["component_manager_id"] = self.component_manager_id

    st = ET.SubElement(nd, "{%s}sliver_type" % (GNS.REQUEST.name))
    st.attrib["name"] = self.type

    if self.disk_image:
      # TODO: Force disk images to be objects, and stop supporting old style strings
      if isinstance(self.disk_image, (str, unicode)):
        di = ET.SubElement(st, "{%s}disk_image" % (GNS.REQUEST.name))
        di.attrib["name"] = self.disk_image
      elif isinstance(self.disk_image, geni.urn.Base):
        di = ET.SubElement(st, "{%s}disk_image" % (GNS.REQUEST.name))
        di.attrib["name"] = str(self.disk_image)
      else:
        self.disk_image._write(st)

    if self.hardware_type:
      hwt = ET.SubElement(nd, "{%s}hardware_type" % (GNS.REQUEST.name))
      hwt.attrib["name"] = self.hardware_type

    if self.interfaces:
      for intf in self.interfaces:
        intf._write(nd)

    if self.services:
      svc = ET.SubElement(nd, "{%s}services" % (GNS.REQUEST.name))
      for service in self.services:
        service._write(svc)

    if self.routable_control_ip:
      ET.SubElement(nd, "{%s}routable_control_ip" % (Namespaces.EMULAB.name))

    for obj in self._ext_children:
      obj._write(nd)

    for elem in self._raw_elements:
      nd.append(elem)

    return nd

  def addInterface (self, name = None, address = None):
    existingNames = [x.name for x in self.interfaces]
    if name is not None:
      if name.find(":") > 0:
        intfName = name
      else:
        intfName = "%s:%s" % (self.client_id, name)
        pass
    else:
      for i in range(0, 100):
        intfName = "%s:if%i" % (self.client_id, i)
        if intfName not in existingNames:
          break

    if intfName in existingNames:
      raise Node.DuplicateInterfaceName()

    intf = Interface(intfName, self, address)
    self.interfaces.append(intf)
    return intf

  def addService (self, svc):
    self.services.append(svc)

  def addRawElement (self, elem):
    self._raw_elements.append(elem)

Request.EXTENSIONS.append(("Node", Node))

class RawPC(Node):
  def __init__ (self, name, component_id = None):
    super(RawPC, self).__init__(name, NodeType.RAW, component_id = component_id, exclusive = True)

Request.EXTENSIONS.append(("RawPC", RawPC))

class VZContainer(Node):
  def __init__ (self, name, exclusive = False):
    super(VZContainer, self).__init__(name, "emulab-openvz", exclusive)



class Namespaces(object):
  CLIENT = GNS.Namespace("client", "http://www.protogeni.net/resources/rspec/ext/client/1")
  RS = GNS.Namespace("rs", "http://www.protogeni.net/resources/rspec/ext/emulab/1")
  EMULAB = GNS.Namespace("emulab", "http://www.protogeni.net/resources/rspec/ext/emulab/1")
  VTOP  = GNS.Namespace("vtop", "http://www.protogeni.net/resources/rspec/ext/emulab/1", "vtop_extension.xsd")
  TOUR =  GNS.Namespace("tour", "http://www.protogeni.net/resources/rspec/ext/apt-tour/1")
  JACKS = GNS.Namespace("jacks", "http://www.protogeni.net/resources/rspec/ext/jacks/1")
  INFO = GNS.Namespace("info", "http://www.protogeni.net/resources/rspec/ext/site-info/1")
  PARAMS = GNS.Namespace("parameters", "http://www.protogeni.net/resources/rspec/ext/profile-parameters/1")
  DATA = GNS.Namespace("data", "http://www.protogeni.net/resources/rspec/ext/user-data/1")
  DELAY =  GNS.Namespace("delay", "http://www.protogeni.net/resources/rspec/ext/delay/1")


#### DEPRECATED #####
class XenVM(Node):
  """
.. deprecated:: 0.4
   Use :py:class:`geni.rspec.igext.XenVM` instead."""
  def __init__ (self, name, component_id = None, exclusive = False):
    import geni.warnings as GW
    import warnings
    warnings.warn("geni.rspec.pg.XenVM is deprecated, please use geni.rspec.igext.XenVM instead",
                  GW.GENILibDeprecationWarning)
    super(XenVM, self).__init__(name, NodeType.XEN, component_id = component_id, exclusive = exclusive)
    self.cores = 1
    self.ram = 256
    self.disk = 8

  def _write (self, root):
    nd = super(XenVM, self)._write(root)
    st = nd.find("{%s}sliver_type" % (GNS.REQUEST.name))
    xen = ET.SubElement(st, "{%s}xen" % (Namespaces.EMULAB.name))
    xen.attrib["cores"] = str(self.cores)
    xen.attrib["ram"] = str(self.ram)
    xen.attrib["disk"] = str(self.disk)
    return nd

VM = XenVM



# Copyright (c) 2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import datetime

from lxml import etree as ET

from . import pg
from .. import namespaces
from ..model.util import XPathXRange

STITCHNS = namespaces.Namespace("stitch", "http://hpn.east.isi.edu/rspec/ext/stitch/0.1/")

_XPNS = {'t' : STITCHNS.name}

class StitchInfo(pg.Resource):
  def __init__ (self):
    super(StitchInfo, self).__init__()
    self._paths = []
    self.addNamespace(STITCHNS)

  def addPath (self, path):
    self._paths.append(path)
    return path

  def _write (self, element):
    se = ET.SubElement(element, "{%s}stitching" % (STITCHNS))
    se.attrib["lastUpdateTime"] = datetime.datetime.now().strftime("%Y%m%d:%H:%M:%S")

    for path in self._paths:
      path._write(se)

    return se


class Path(object):
  def __init__ (self, name):
    self.name = name
    self._hops = []

  def addHop (self, hop):
    self._hops.append(hop)
    return hop

  def _write (self, element):
    pe = ET.SubElement(element, "{%s}path" % (STITCHNS))
    pe.attrib["id"] = self.name

    # We wait until writing to give them IDs, so you can do all kinds of stupid things
    # like reordering and deleting until then
    for idx,hop in enumerate(self._hops, start=1):
      hop._id = idx
      if len(self._hops) != idx:
        hop._next_hop_id = idx+1

    for hop in self._hops:
      hop._write(pe)

    return pe


class Hop(object):
  def __init__ (self):
    self.link_id = None
    self.capacity = 1
    self.suggested_vlan = None
    self._id = None
    self._next_hop_id = None

    # Don't override these unless the aggregate hates you
    self.ad_vrange_low = 1
    self.ad_vrange_high = 4092
    self.vlan_translation = False

  def _write (self, element):
    he = ET.SubElement(element, "{%s}hop" % (STITCHNS))
    he.attrib["id"] = "%d" % (self._id)

    link = ET.SubElement(he, "{%s}link" % (STITCHNS))
    link.attrib["id"] = self.link_id

    tem = ET.SubElement(link, "{%s}trafficEngineeringMetric" % (STITCHNS))
    tem.text = "10"

    ce = ET.SubElement(link, "{%s}capacity" % (STITCHNS))
    ce.text = "%d" % (self.capacity)

    scd = ET.SubElement(link, "{%s}switchingCapabilityDescriptor" % (STITCHNS))

    sct = ET.SubElement(scd, "{%s}switchingcapType" % (STITCHNS))
    sct.text = "l2sc"

    enc = ET.SubElement(scd, "{%s}encodingType" % (STITCHNS))
    enc.text = "ethernet"

    scs = ET.SubElement(scd, "{%s}switchingCapabilitySpecificInfo" % (STITCHNS))
    l2scs = ET.SubElement(scs, "{%s}switchingCapabilitySpecificInfo_L2sc" % (STITCHNS))

    imtu = ET.SubElement(l2scs, "{%s}interfaceMTU" % (STITCHNS))
    imtu.text = "1500"

    vra = ET.SubElement(l2scs, "{%s}vlanRangeAvailability" % (STITCHNS))
    vra.text = "%d-%d" % (self.ad_vrange_low, self.ad_vrange_high)

    svr = ET.SubElement(l2scs, "{%s}suggestedVLANRange" % (STITCHNS))
    svr.text = "%d" % (self.suggested_vlan)

    vt = ET.SubElement(l2scs, "{%s}vlanTranslation" % (STITCHNS))
    vt.text = str(self.vlan_translation).lower()

    nh = ET.SubElement(he, "{%s}nextHop" % (STITCHNS))
    if self._next_hop_id:
      nh.text = "%d" % (self._next_hop_id)
    else:
      nh.text = "null"

    return he

def coerceBool (text):
  if text.lower() == "false":
    return False
  elif text.lower() == "true":
    return True
  else:
    return None

class AdInfo(object):
  def __init__ (self, elem):
    self._root = elem
    self._aggregates = {}

  @property
  def aggregates (self):
    if not self._aggregates:
      for elem in self._root.xpath('t:aggregate', namespaces=_XPNS):
        info = AggInfo(elem)
        self._aggregates[info.urn] = info
    return self._aggregates


class AggInfo(object):
  def __init__ (self, elem):
    self._root = elem
    self.urn = elem.get("id")
    self.url = elem.get("url")

  @property
  def mode (self):
    return self._root.xpath("t:stitchingmode", namespaces=_XPNS)[0].text

  @property
  def scheduledservices (self):
    t = self._root.xpath("t:scheduledservices", namespaces=_XPNS)[0].text
    return coerceBool(t)

  @property
  def negotiatedservices (self):
    t = self._root.xpath("t:negotiatedservices", namespaces=_XPNS)[0].text
    return coerceBool(t)

  @property
  def nodes (self):
    n = self._root.xpath("t:node", namespaces=_XPNS)
    return XPathXRange(n, AggNode)


class AggNode(object):
  def __init__ (self):
    self.id = None
    self._root = None

  @property
  def ports (self):
    p = self._root.xpath("t:port", namespaces=_XPNS)
    return XPathXRange(p, AggPort)

  @classmethod
  def _fromdom (cls, elem):
    node = AggNode()
    node._root = elem
    node.id = elem.get("id")
    return node


class AggPort(object):
  def __init__ (self):
    self.id = None
    self.capacity = 0
    self._root = None

  @property
  def links (self):
    l = self._root.xpath("t:link", namespaces=_XPNS)
    return XPathXRange(l, AggLink)

  @classmethod
  def _fromdom (cls, elem):
    port = AggPort()
    port._root = elem
    port.id = elem.get("id")
    port.capacity = int((elem.xpath("t:capacity", namespaces=_XPNS)[0].text).strip("kbps"))
    return port


class AggLink(object):
  def __init__ (self):
    self._root = None
    self.id = None
    self.remote_urn = None

  @property
  def al2sinfo (self):
    if not self.remote_urn.count("al2s.internet2.edu"):
      return None
    parts = self.remote_urn.split(":")
    port = parts[-2]
    switch = parts[-3].split("+")[-1]
    return (switch, port)

  @classmethod
  def _fromdom (cls, elem):
    link = AggLink()
    link._root = elem
    link.id = elem.get("id")
    link.remote_urn = elem.xpath("t:remoteLinkId", namespaces=_XPNS)[0].text
    return link



# Copyright (c) 2013-2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

class Namespace(object):
  def __init__ (self, prefix, name, location = None):
    self.prefix = prefix
    self.name = name
    self.location = location

  def __repr__ (self):
    return self.name

XSNS = Namespace("xsi", "http://www.w3.org/2001/XMLSchema-instance")

REQUEST = Namespace("request", "http://www.geni.net/resources/rspec/3", "http://www.geni.net/resources/rspec/3/request.xsd")
OFv3 = Namespace("openflow", "http://www.geni.net/resources/rspec/ext/openflow/3")
OFv4 = Namespace("openflow", "http://www.geni.net/resources/rspec/ext/openflow/4")
SVLAN = Namespace("sharedvlan", "http://www.geni.net/resources/rspec/ext/shared-vlan/1",
                  "http://www.geni.net/resources/rspec/ext/shared-vlan/1/request.xsd")
OPSTATE = Namespace("opstate", "http://www.geni.net/resources/rspec/ext/opstate/1",
                    "http://http://www.geni.net/resources/rspec/ext/opstate/1/ad.xsd")
USER = Namespace("user", "http://www.geni.net/resources/rspec/ext/user/1")



# Copyright (c) 2014-2016  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import re
import sys
import inspect

from lxml import etree as ET

from .. import namespaces as GNS
from .pg import Namespaces as PGNS
from .pg import Node
from .pg import Link
from .pg import Resource
from . import pg
from .. import urn

class OFController(object):
  """OpenFlow controller specification to be used on a PG VLAN.

Add to link objects using the Link.addChild() method.

.. note::
  This will have no effect if a trivial link is created by the aggregate.
  You need to make sure that a VLAN will be provisioned (typically by making sure
  that at least two interfaces on the link are on different physical hosts)."""

  def __init__ (self, host, port=6633):
    self.host = host
    self.port = port

  def _write (self, element):
    eof = ET.SubElement(element, "{%s}openflow_controller" % (PGNS.EMULAB))
    eof.attrib["url"] = "tcp:%s:%d" % (self.host, self.port)
    return eof


class XenVM(Node):
  """Xen-based Virtual Machine resource

  Args:
    client_id (str): Your name for this VM.  This must be unique within a single `Request` object.
    component_id (Optional[str]): The `component_id` of the site node you want to bind this VM to
    exclusive (Optional[bool]): Request this VM on an isolated host used only by your sliver.

  Attributes:
    cores (int): Number of CPU cores
    ram (int): Amount of memory in megabytes
    disk (int): Amount of disk space in gigabytes
    xen_ptype (str): Physical node type on which to instantiate the VM. Types are AM-specific.
  """
  def __init__ (self, client_id, component_id = None, exclusive = False):
    super(XenVM, self).__init__(client_id, "emulab-xen", component_id = component_id, exclusive = exclusive)
    self.cores = None
    self.ram = None
    self.disk = None
    self.xen_ptype = None

  def _write (self, root):
    nd = super(XenVM, self)._write(root)
    st = nd.find("{%s}sliver_type" % (GNS.REQUEST.name))
    if self.cores or self.ram or self.disk:
      xen = ET.SubElement(st, "{%s}xen" % (PGNS.EMULAB.name))
      if self.cores:
        xen.attrib["cores"] = str(self.cores)
      if self.ram:
        xen.attrib["ram"] = str(self.ram)
      if self.disk:
        xen.attrib["disk"] = str(self.disk)
    if self.xen_ptype is not None:
      pt = ET.SubElement(st, "{%s}xen_ptype" % (PGNS.EMULAB.name))
      pt.attrib["name"] = self.xen_ptype
    return nd

pg.Request.EXTENSIONS.append(("XenVM", XenVM))


class AddressPool(Resource):
  """A pool of public dynamic IP addresses belonging to a slice."""

  def __init__(self, name, count=1, type="any"):
    super(AddressPool, self).__init__()
    self.client_id = name
    self.count = count
    self.type = type
    self.component_manager_id = None

  @property
  def name (self):
    return self.client_id

  def _write (self, root):
    pl = ET.SubElement(root, "{%s}routable_pool" % (PGNS.EMULAB.name))
    pl.attrib["client_id"] = self.client_id
    if self.component_manager_id:
      pl.attrib["component_manager_id"] = self.component_manager_id

    pl.attrib["count"] = str(self.count)
    pl.attrib["type"] = self.type

    return pl

pg.Request.EXTENSIONS.append(("AddressPool", AddressPool))


class Blockstore(object):
  def __init__ (self, name, mount = None):
    """Creates a BlockStore object with the given name (arbitrary) and mountpoint."""
    self.name = name
    self.mount = mount
    self._size = None
    self.where = "local"    # local|remote
    self.readonly = False
    self.placement = "any"  # any|sysvol|nonsysvol
    self.dataset = None
    self.rwclone = False    # Only for remote blockstores.

  @property
  def size (self):
      return self._size

  @size.setter
  def size (self, val):
      match = re.match(r"^(\d+)GB$", val)
      if match:
          self._size = int(match.group(1))
      else:
          self._size = int(val)

  def _write (self, element):
    bse = ET.SubElement(element, "{%s}blockstore" % (PGNS.EMULAB))
    bse.attrib["name"] = self.name
    if self.mount:
      bse.attrib["mountpoint"] = self.mount
    bse.attrib["class"] = self.where
    if self._size:
      bse.attrib["size"] = "%dGB" % (self._size)
    bse.attrib["placement"] = self.placement
    if self.readonly:
      bse.attrib["readonly"] = "true"
    if self.rwclone:
      bse.attrib["rwclone"] = "true"
    if self.dataset:
      if isinstance(self.dataset, (str, unicode)):
        bse.attrib["dataset"] = self.dataset
      elif isinstance(self.dataset, urn.Base):
        bse.attrib["dataset"] = str(self.dataset)
    return bse

pg.Node.EXTENSIONS.append(("Blockstore", Blockstore))


class RemoteBlockstore(pg.Node):
  def __init__ (self, name, mount = None, ifacename = "if0"):
    super(RemoteBlockstore, self).__init__(name, "emulab-blockstore")
    bs = Blockstore(self.name, mount)
    bs.where = "remote"
    self._bs = bs
    self._interface = self.addInterface(ifacename)

  def _write (self, element):
    return self._bs._write(super(RemoteBlockstore, self)._write(element));

  @property
  def interface (self):
    return self._interface

  @property
  def size (self):
    return self._bs.size

  @size.setter
  def size (self, val):
    self._bs.size = val

  @property
  def mountpoint (self):
    return self._bs.mount

  @mountpoint.setter
  def mountpoint (self, val):
    self._bs.mount = val

  @property
  def readonly (self):
    return self._bs.readonly

  @readonly.setter
  def readonly (self, val):
    self._bs.readonly = val

  @property
  def rwclone (self):
    return self._bs.rwclone

  @rwclone.setter
  def rwclone (self, val):
    self._bs.rwclone = val

  @property
  def placement (self):
    return self._bs.placement

  @placement.setter
  def placement (self, val):
    self._bs.placement = val

  @property
  def dataset (self):
    return self._bs.dataset

  @dataset.setter
  def dataset (self, val):
    self._bs.dataset = val

pg.Request.EXTENSIONS.append(("RemoteBlockstore", RemoteBlockstore))

class Bridge(pg.Node):
  class Pipe(object):
    def __init__ (self):
      self.bandwidth = 0
      self.latency   = 0
      self.lossrate  = 0.0
  
  def __init__ (self, name, if0name = "if0", if1name = "if1"):
    super(Bridge, self).__init__(name, "delay")
    self.addNamespace(PGNS.DELAY)
    
    self.iface0 = self.addInterface(if0name)
    self.pipe0  = self.Pipe();
    self.iface1 = self.addInterface(if1name)
    self.pipe1  = self.Pipe();

  def _write (self, root):
    nd = super(Bridge, self)._write(root)
    st = nd.find("{%s}sliver_type" % (GNS.REQUEST.name))
    delay = ET.SubElement(st, "{%s}sliver_type_shaping" % (PGNS.DELAY.name))
    pipe0 = ET.SubElement(delay, "{%s}pipe" % (PGNS.DELAY.name))
    pipe0.attrib["source"]    = self.iface0.name
    pipe0.attrib["dest"]      = self.iface1.name
    pipe0.attrib["capacity"]  = str(self.pipe0.bandwidth)
    pipe0.attrib["latency"]   = str(self.pipe0.latency)
    pipe0.attrib["lossrate"]  = str(self.pipe0.lossrate)
    pipe1 = ET.SubElement(delay, "{%s}pipe" % (PGNS.DELAY.name))
    pipe1.attrib["source"]    = self.iface1.name
    pipe1.attrib["dest"]      = self.iface0.name
    pipe1.attrib["capacity"]  = str(self.pipe1.bandwidth)
    pipe1.attrib["latency"]   = str(self.pipe1.latency)
    pipe1.attrib["lossrate"]  = str(self.pipe1.lossrate)
    return nd;

  # pipe0 goes with iface0, and pipe1 goes with iface1
  def getPipe (self, interface):
    if self.iface0.name is interface: 
      return self.pipe0
    else:
      return self.pipe1
  
pg.Request.EXTENSIONS.append(("Bridge", Bridge))

class Firewall(object):
  class Style(object):
    OPEN     = "open"
    CLOSED   = "closed"
    BASIC    = "basic"

  class Direction(object):
    INCOMING = "incoming"
    OUTGOING = "outgoing"

  def __init__ (self, style):
    self.style = style
    self.exceptions = []

  def addException(self, port, direction, ip = None):
    self.exceptions.append({"port" : port, "direction" : direction, "ip" : ip})

  def _write (self, node):
    fw = ET.SubElement(node, "{%s}firewall" % (PGNS.EMULAB))
    fw.attrib["style"] = self.style
    for excep in self.exceptions:
      ex = ET.SubElement(fw, "exception")
      ex.attrib["port"]      = str(excep["port"])
      ex.attrib["direction"] = excep["direction"]
      if excep["ip"]:
        ex.attrib["ip"] = excep["ip"]
    return fw

XenVM.EXTENSIONS.append(("Firewall", Firewall))


class Tour(object):
  TEXT = "text"
  MARKDOWN = "markdown"

  # One or more blank lines, followed by "Instructions:" on it's own line, then
  # zero or more blank lines. Eats the blank lines.
  SPLIT_REGEX = re.compile(r"\n+^\w*instructions\w*:?\w*$\n+",
                           re.IGNORECASE | re.MULTILINE)

  class Step(object):
    # Duplicated because of the awkwardness of accessing class variables in
    # outer class
    TEXT = "text"
    MARKDOWN = "markdown"
    def __init__(self, target, description,
        steptype = None, description_type = MARKDOWN):
      if hasattr(target,'client_id'):
        self.id = target.client_id
      else:
        self.id = str(target)

      if steptype:
        self.type = steptype
      elif isinstance(target, Node):
        self.type = "node"
      elif isinstance(target, Link):
        self.type = "link"
      else:
        self.type = "node"

      self.description = description
      self.description_type = description_type

    def _write (self, root):
      stepel = ET.SubElement(root, "step")
      stepel.attrib["point_type"] = self.type
      stepel.attrib["point_id"]   = self.id
      desc = ET.SubElement(stepel, "description")
      desc.text = self.description
      desc.attrib["type"] = self.description_type

  def __init__ (self):
    self.description = None
    # Type can markdown
    self.description_type = Tour.TEXT
    self.instructions = None
    # Type can markdown
    self.instructions_type = Tour.TEXT
    self.steps = []

  def addStep(self, step):
    self.steps.append(step)

  def Description(self, type, desc):
    self.description_type = type
    self.description = desc

  def Instructions(self, type, inst):
    self.instructions_type = type
    self.instructions = inst

  def useDocstring(self, module = None):
    if module is None:
      module = sys.modules["__main__"]
    if not self.description and module.__doc__:
      docstr = module.__doc__
      docparts = Tour.SPLIT_REGEX.split(docstr,2)
      self.Description(Tour.MARKDOWN,docparts[0])
      if len(docparts) == 2 and not self.instructions:
        self.Instructions(Tour.MARKDOWN,docparts[1])
      return True
    else:
      return False

  def _write (self, root):
    #
    # Please do it this way, until some of our JS code is fixed.
    #
    td = ET.SubElement(root, "rspec_tour",
                       nsmap={None : PGNS.TOUR.name})
    if self.description:
      desc = ET.SubElement(td, "description")
      desc.text = self.description
      desc.attrib["type"] = self.description_type
    if self.instructions:
      inst = ET.SubElement(td, "instructions")
      inst.text = self.instructions
      inst.attrib["type"] = self.instructions_type
    if len(self.steps):
      steps = ET.SubElement(td, "steps")
      for step in self.steps:
        step._write(steps)
    return td


class ParameterData(object):
  def __init__ (self, parameters):
    self.parameters = parameters

  def _write (self, root):
    td = ET.SubElement(root, "data_set",
                       nsmap={None : PGNS.PARAMS.name})
    for param in self.parameters:
      that = self.parameters[param]
      if 'hide' in that and that['hide'] is False:
        desc = ET.SubElement(td, "data_item")
        desc.attrib["name"] = that['prefix'] + param

        if isinstance(that['value'], ET._Element):
          desc.append(that['value'])
        else:
          desc.text = str(that['value'])

pg.Request.EXTENSIONS.append(("ParameterData", ParameterData))


class Site(object):
  def __init__ (self, id):
    self.id = id

  def _write (self, node):
    site = ET.SubElement(node, "{%s}site" % (PGNS.JACKS))
    site.attrib["id"] = self.id
    return site

pg.Node.EXTENSIONS.append(("Site", Site))
pg.Link.EXTENSIONS.append(("Site", Site))


class Desire(object):
  def __init__ (self, name, weight):
    self.name = name
    self.weight = weight

  def _write (self, node):
    fd = ET.SubElement(node, "{%s}fd" % (PGNS.EMULAB))
    fd.attrib["name"] = self.name
    fd.attrib["weight"] = str(self.weight)
    return fd

pg.Node.EXTENSIONS.append(("Desire", Desire))


class Password(Resource):
  """A declaration for a randomly generated password.

The portal will generate the password, encrypt it, and pass on the
encrypted value to the AM(s) and therefore the node(s)."""

  def __init__(self, name=None):
    super(Password, self).__init__()
    self.name = name

  def _write (self, root):
    pl = ET.SubElement(root, "{%s}password" % (PGNS.EMULAB.name))
    if self.name:
      pl.attrib["name"] = self.name

    return pl



# Copyright (c) 2014-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import abc

from .core import AMTypeRegistry

class AMType(object):
  __metaclass__ = abc.ABCMeta

  def __init__ (self, name):
    self.name = name

  @abc.abstractmethod
  def parseAdvertisement (self, data):
    return

  @abc.abstractmethod
  def parseManifest (self, data):
    return


class ExoGENI(AMType):
  def __init__ (self, name="exogeni"):
    super(ExoGENI, self).__init__ (name)

  def parseAdvertisement (self, data):
    from ..rspec import pgad
    ad = pgad.Advertisement(xml=data["value"])
    return ad

  def parseManifest (self, data):
    from ..rspec import pgmanifest
    manifest = pgmanifest.Manifest(xml = data["value"])
    return manifest


class ProtoGENI(AMType):
  def __init__ (self, name="pg"):
    super(ProtoGENI, self).__init__(name)

  def parseAdvertisement (self, data):
    from ..rspec import pgad
    ad = pgad.Advertisement(xml=data["value"])
    ad.error_url = data["code"]["protogeni_error_url"]
    return ad

  def parseManifest (self, data):
    from ..rspec import pgmanifest
    if isinstance(data, (str, unicode)):
      manifest = pgmanifest.Manifest(xml = data)
    else:
      manifest = pgmanifest.Manifest(xml = data["value"])
      manifest.error_url = data["code"]["protogeni_error_url"]
    return manifest


class FOAM(AMType):
  def __init__ (self, name="foam"):
    super(FOAM, self).__init__(name)

  def parseAdvertisement (self, data):
    from ..rspec import ofad
    ad = ofad.Advertisement(xml=data["value"])
    return ad

  def parseManifest (self, data):
    from ..rspec import ofmanifest
    manifest = ofmanifest.Manifest(xml = data["value"])
    return manifest


class OpenGENI(AMType):
  def __init__ (self, name="opengeni"):
    super(OpenGENI, self).__init__(name)

  def parseAdvertisement (self, data):
    from ..rspec import pgad
    ad = pgad.Advertisement(xml=data["value"])
    return ad

  def parseManifest (self, data):
    from ..rspec import pgmanifest
    manifest = pgmanifest.Manifest(xml = data["value"])
    return manifest


class VTS(AMType):
  def __init__ (self, name="vts"):
    super(VTS, self).__init__(name)

  def parseAdvertisement (self, data):
    from ..rspec import vtsad
    ad = vtsad.Advertisement(xml=data["value"])
    return ad

  def parseManifest (self, data):
    from ..rspec import vtsmanifest
    if isinstance(data, (str, unicode)):
      manifest = vtsmanifest.Manifest(xml = data)
    else:
      manifest = vtsmanifest.Manifest(xml = data["value"])
    return manifest


class OESS(AMType):
  def __init__ (self, name="oess"):
    super(OESS, self).__init__(name)

  def parseAdvertisement (self, data):
    from ..rspec import oessad
    ad = oessad.Advertisement(xml=data["value"])
    return ad

  def parseManifest (self, data):
    from ..rspec import oessmanifest
    if isinstance(data, (str, unicode)):
      manifest = oessmanifest.Manifest(xml = data)
    else:
      manifest = oessmanifest.Manifest(xml = data["value"])
    return manifest


AMTypeRegistry.register("foam", FOAM())
AMTypeRegistry.register("opengeni", OpenGENI())
AMTypeRegistry.register("pg", ProtoGENI())
AMTypeRegistry.register("exogeni", ExoGENI())
AMTypeRegistry.register("vts", VTS())
AMTypeRegistry.register("oess", OESS())



# Copyright (c) 2014-2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import sys
import inspect

from .core import AM

class OF(AM):
  def __init__ (self, name, host, url = None):
    if url is None:
      url = "https://%s:3626/foam/gapi/2" % (host)
    super(OF, self).__init__(name, url, "amapiv2", "foam")


Internet2 = OF("of-i2", "foam.net.internet2.edu")
UEN = OF("of-uen", "foamyflow.chpc.utah.edu")


def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      yield obj



# Copyright (c) 2014  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import inspect
import sys

from .core import AM

class EGCompute(AM):
  def __init__ (self, name, host, cmid = None, url = None):
    if url is None:
      url = "https://%s:11443/orca/xmlrpc" % (host)
    super(EGCompute, self).__init__(name, url, "amapiv2", "exogeni", cmid)

EXOSM = EGCompute("exosm", "geni.renci.org")
GPO = EGCompute("eg-gpo", "bbn-hn.exogeni.net")
RCI = EGCompute("eg-rci", "rci-hn.exogeni.net")
FIU = EGCompute("eg-fiu", "fiu-hn.exogeni.net")
UH = EGCompute("eg-uh", "uh-hn.exogeni.net")
NCSU = EGCompute("eg-ncsu", "ncsu-hn.exogeni.net")
UFL = EGCompute("eg-ufl", "ufl-hn.exogeni.net")
OSF = EGCompute("eg-osf", "osf-hn.exogeni.net")
NICTA = EGCompute("eg-nicta", "nicta-hn.exogeni.net")
SL = EGCompute("eg-sl", "sl-hn.exogeni.net")
TAMU = EGCompute("eg-tamu", "tamu-hn.exogeni.net")
WVN = EGCompute("eg-wvn", "wvn-hn.exogeni.net")
WSU = EGCompute("eg-wsu", "wsu-hn.exogeni.net")

def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      yield obj

def name_to_aggregate ():
  result = dict()
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      result[obj.name] = obj
  return result



# Copyright (c) 2014  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

class XPathXRange(object):
  def __init__ (self, xp, klass):
    self._data = xp
    self._klass = klass
  def __iter__ (self):
    for obj in self._data:
      yield self._klass._fromdom(obj)
  def __len__ (self):
    return len(self._data)
  def __getitem__ (self, idx):
    return self._klass._fromdom(self._data[idx])



# Copyright (c) 2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import requests
import requests.auth

from ..types import DPID


class Workgroup(object):
  def __init__ (self, name, workgroup_id):
    self.workgroup_id = int(workgroup_id)
    self.name = name

  def __repr__ (self):
    return "[%d] %s" % (self.workgroup_id, self.name)

class Node(object):
  def __init__ (self, ndata):
    self.dpid = DPID(long(ndata["dpid"]))
    self.name = ndata["name"]
    self.node_id = int(ndata["node_id"])
    self.state = ndata["operational_state"]
    self.send_barrier_bulk = bool(int(ndata["send_barrier_bulk"]))

  def __repr__ (self):
    return str(self.__dict__)

class Connection(object):
  BASE = "https://%s/oess/services/%s"

  def __init__ (self, hostname, uname, passwd):
    self.hostname = hostname
    self.url = Connection.BASE % (self.hostname, "data.cgi")
    self.purl = Connection.BASE % (self.hostname, "provisioning.cgi")
    self.auth = requests.auth.HTTPBasicAuth(uname, passwd)
    self.workgroup_id = None

  def setAuthInfo (self, uname, passwd):
    self.auth = requests.auth.HTTPBasicAuth(uname, passwd)

  @property
  def workgroups (self):
    r = requests.get(self.url, auth=self.auth, params={"action":"get_workgroups"})
    wgs = []
    for wg in r.json()["results"]:
      wgs.append(Workgroup(wg["name"], wg["workgroup_id"]))
    return wgs

  @property
  def nodes (self):
    r = requests.get(self.url, auth=self.auth, params={"action":"get_all_node_status"})
    nl = []
    for nd in r.json()["results"]:
      nl.append(Node(nd))
    return nl

  @property
  def circuits (self):
    r = requests.get(self.url, auth=self.auth, params={"action":"get_existing_circuits",
                                                       "workgroup_id":self.workgroup_id})
    return r.json()["results"]

  def getNodeInterfaces (self, name):
    r = requests.get(self.url, auth=self.auth,
                     params={"action":"get_node_interfaces", "workgroup_id":self.workgroup_id,
                             "node":name})
    return r.json()["results"]

  def provisionVLANPatch (self, desc, tag, node_name, intf_a_name, intf_b_name):
    payload = {"action" : "provision_circuit", "workgroup_id" : self.workgroup_id,
               "circuit_id" : -1, "description" : desc, "bandwidth" : 0, "provision_time": -1,
               "remove_time" : -1, "link" : [], "backup_link" : [], "remote_nodes" : [],
               "remote_tags" : [], "restore_to_primary" : 0, "node" : [], "interface" : [],
               "tag" : []}

    payload["node"].append(node_name)
    payload["interface"].append(intf_a_name)
    payload["tag"].append(tag)

    payload["node"].append(node_name)
    payload["interface"].append(intf_b_name)
    payload["tag"].append(tag)

    r = requests.post(self.purl, auth=self.auth, data = payload)
    return r.json()["results"]



# Copyright (c) 2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import requests

import geni._coreutil as GCU

import requests.packages.urllib3
import requests.packages.urllib3.exceptions
requests.packages.urllib3.disable_warnings((requests.packages.urllib3.exceptions.InsecureRequestWarning,
                                            requests.packages.urllib3.exceptions.InsecurePlatformWarning))

class Link(object):
  def __init__ (self, item):
    self.rel = item["rel"]
    self.href = item["href"]
    self.type = item["type"]


class ResourceDiscovery(object):
  def __init__ (self, name, host, port = 443, url = None):
    self.name = name
    self.url = url
    if self.url is None:
      self.url = "https://%s:%d" % (host, port)

    self._uid = None
    self._version = None
    self._release = None
    self._links = {}

  def _populate (self):
    r = requests.get(self.url, **self.rkwargs)
    data = r.json()
    self._uid = data["uid"]
    self._version = data["version"]
    self._release = data["release"]
    for item in data["links"]:
      self._links[item["rel"]] = Link(item)

  @property
  def rkwargs (self):
    d = {"headers" : self.headers,
         "verify" : False}
    return d

  @property
  def headers (self):
    d = {"Content-Type" : "application/json"}
    d.update(GCU.defaultHeaders())
    return d

  @property
  def uid (self):
    if not self._uid:
      self._populate()
    return self._uid

  @property
  def version (self):
    if not self._version:
      self._populate()
    return self._version
    
  @property
  def release (self):
    if not self._release:
      self._populate()
    return self._release

  @property
  def sites (self):
    if not self._links:
      self._populate()
    shref = self._links["sites"].href
    r = requests.get("%s%s" % (self.url, shref), **self.rkwargs)
    data = r.json()
    for site in data["items"]:
      yield Site(site)


class Site(object):
  def __init__ (self, item):
    self._links = {}
    self.description = item["description"]
    self.name = item["name"]
    self.uid = item["uid"]
    self.location = item["location"]
    self.latitude = item["latitude"]
    self.longitude = item["longitude"]
    for ldata in item["links"]:
      self._links[ldata["rel"]] = Link(ldata)




# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import sys
import inspect

from .core import AM, APIRegistry

class HostPOAs(object):
  def __init__ (self, vtsam):
    self.am = vtsam

  def getARPTable (self, context, sname, client_ids):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self.am._apiv3.poa(context, self.am.urlv3, sname, "api:uh.host:get-arp-table",
                              options={"client-ids": client_ids})

  def getRouteTable (self, context, sname, client_ids):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self.am._apiv3.poa(context, self.am.urlv3, sname, "api:uh.host:get-route-table",
                              options={"client-ids": client_ids})

  def svcStatus (self, context, sname, client_ids):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self.am._apiv3.poa(context, self.am.urlv3, sname, "api:uh.host:supervisor-status",
                              options={"client-ids": client_ids})

  def execcmd (self, context, sname, client_ids, cmd):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self.am._apiv3.poa(context, self.am.urlv3, sname, "api:uh.host:exec",
                              options={"client-ids": client_ids, "cmd" : cmd})


class v4RouterPOAs(object):
  def __init__ (self, vtsam):
    self.am = vtsam

  def addOSPFNetworks (self, context, sname, client_ids, nets):
    """Add OSPF Networks to areas on the given routers

    Args:
      context: geni-lib context
      sname (str): Slice name
      client_ids (list): A list of client-id strings
      nets (list): A list of (network, area) tuples
    """
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self.am._apiv3.poa(context, self.am.urlv3, sname, "vts:uh.quagga:add-ospf-nets",
                              options={"client-ids": client_ids, "networks" : nets})

  def getRouteTable (self, context, sname, client_ids):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self.am._apiv3.poa(context, self.am.urlv3, sname, "vts:uh.quagga:get-route-table",
                              options={"client-ids": client_ids})

  def getOSPFNeighbors (self, context, sname, client_ids):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self.am._apiv3.poa(context, self.am.urlv3, sname, "vts:uh.quagga:get-ospf-neighbors",
                              options={"client-ids": client_ids})


class Policy(object):
  def __init__ (self, vtsam):
    self.am = vtsam

  # Policy consent hooks for GDPR-style compliance
  def getText (self, context, pid = None):
    """Get the text contents of the policy requested.  If no policy is specified and only one policy
    exists at the aggregate, that policy text will be returned.

    Args:
      context: geni-lib context
      pid: policy ID (typically from `getversion` output)

    Returns:
      str: Text contents of policy
    """
    opts = {}
    if pid:
      opts["policy-id"] = pid
    return self.am._apiv3.paa(context, self.am.urlv3, "vts:policy:get-text", options = opts)

  def giveConsent (self, context, pid):
    """Give consent to the policy indicated for the user URN in the credential used.

    Args:
      context: geni-lib context
      pid: policy ID
    """
    return self.am._apiv3.paa(context, self.am.urlv3, "vts:policy:consent", options = {"policy-id" : pid})

  def revokeConsent (self, context, pid):
    """Revoke consent from this date forward to the policy indicated for the user URN in the credential used.

    Args:
      context: geni-lib context
      pid: policy ID
    """
    return self.am._apiv3.paa(context, self.am.urlv3, "vts:policy:revoke", options = {"policy-id" : pid})


class VTS(AM):
  """Wrapper for all VTS-supported AMAPI functions"""

  def __init__ (self, name, host, url = None):
    self._host = host
    if url is None:
      url = "https://%s:3626/foam/gapi/2" % (self._host)
    self.urlv3 = "%s3" % (url[:-1])
    self._apiv3 = APIRegistry.get("amapiv3")
    super(VTS, self).__init__(name, url, "amapiv2", "vts")
    self.Host = HostPOAs(self)
    self.IPv4Router = v4RouterPOAs(self)
    self.Policy = Policy(self)

  def allocate (self, context, sname, rspec):
    rspec_data = rspec.toXMLString()
    manifest = self._apiv3.allocate(context, self.urlv3, sname, rspec_data)
    return self.amtype.parseManifest(manifest)

  def provision (self, context, sname):
    udata = []
    for user in context._users:
      data = {"urn" : user.urn, "keys" : [open(x, "rb").read() for x in user._keys]}
      udata.append(data)

    res = self._apiv3.provision(context, self.urlv3, sname, options = {"geni_users" : udata})
    if res["code"]["geni_code"] == 0:
      return self.amtype.parseManifest(res["value"])

  def changeController (self, context, sname, url, datapaths, ofver=None):
    options={"controller-url" : url, "datapaths" : datapaths}
    if ofver:
      options["openflow-version"] = ofver
    return self._apiv3.poa(context, self.urlv3, sname, "vts:of:change-controller", options)

  def dumpFlows (self, context, sname, datapaths, show_hidden=False):
    """Get the current flows and flow stats from the requested datapaths.
    
    Args:
      context: geni-lib context
      sname (str): Slice name
      datapaths (list): A list of datapath client_id strings
      show_hidden (bool): Show hidden flows (if any)

    Returns:
      dict: Key/Value dictionary of format `{ client_id : [(flow_field, ...), ...] }`
    """
    return self._apiv3.poa(context, self.urlv3, sname, "vts:of:dump-flows",
                           options={"datapaths" : datapaths, "show-hidden" : show_hidden})

  def getL2Table (self, context, sname, client_ids):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self._apiv3.poa(context, self.urlv3, sname, "api:l2-switch:get-l2-table",
                           options={"client-ids" : client_ids})

  def clearL2Table (self, context, sname, client_ids):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self._apiv3.poa(context, self.urlv3, sname, "api:uh.vswitch:clear-l2-table",
                           options={"client-ids" : client_ids})

  def clearFlows (self, context, sname, datapaths):
    """Clear all installed flows from the requested datapaths.

    Args:
      context: geni-lib context
      sname (str): Slice name
      datapaths (list): A list of datapath client_id strings
    """
    return self._apiv3.poa(context, self.urlv3, sname, "vts:of:clear-flows", options={"datapaths" : datapaths})

  def portDown (self, context, sname, client_id):
    return self._apiv3.poa(context, self.urlv3, sname, "vts:port-down",
                           options={"port-client-id" : client_id})

  def portUp (self, context, sname, client_id):
    return self._apiv3.poa(context, self.urlv3, sname, "vts:port-up",
                           options={"port-client-id" : client_id})

  def addFlows (self, context, sname, flows):
    return self._apiv3.poa(context, self.urlv3, sname, "vts:of:add-flows", options={"rules" : flows})

  def getSTPInfo (self, context, sname, datapaths):
    if not isinstance(datapaths, list): datapaths = [datapaths]
    return self._apiv3.poa(context, self.urlv3, sname, "vts:l2:stp-info",
                           options={"datapaths" : datapaths})

  def getRSTPInfo (self, context, sname, datapaths):
    if not isinstance(datapaths, list): datapaths = [datapaths]
    return self._apiv3.poa(context, self.urlv3, sname, "vts:l2:rstp-info",
                           options={"datapaths" : datapaths})

  def getPortInfo (self, context, sname, datapaths):
    if not isinstance(datapaths, list): datapaths = [datapaths]
    return self._apiv3.poa(context, self.urlv3, sname, "vts:raw:get-port-info",
                           options={"datapaths" : datapaths})

  def setPortBehaviour (self, context, sname, port_list):
    port_json_list = []
    for (port,obj) in port_list:
      port_json_list.append((port, obj.__json__()))
    return self._apiv3.poa(context, self.urlv3, sname, "vts:raw:set-port-behaviour",
                           options={"ports" : port_json_list})

  def getLeaseInfo (self, context, sname, client_ids):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self._apiv3.poa(context, self.urlv3, sname, "api:uh.dhcp:get-leases",
                           options = {"client-ids" : client_ids})

  def setPortVLAN (self, context, sname, port_tuples):
    if not isinstance(port_tuples, list): port_tuples = [port_tuples]
    return self._apiv3.poa(context, self.urlv3, sname, "vts:raw:set-vlan",
                           options = {"ports" : port_tuples})

  def setPortTrunk (self, context, sname, port_list):
    if not isinstance(port_list, list): port_list = [port_list]
    return self._apiv3.poa(context, self.urlv3, sname, "vts:raw:set-trunk",
                           options = {"ports" : port_list})

  def addSSHKeys (self, context, sname, client_ids, keys):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    if not isinstance(keys, list): keys = [keys]
    return self._apiv3.poa(context, self.urlv3, sname, "vts:container:add-keys",
                           options = {"client-ids" : client_ids, "ssh-keys" : keys})

  def setDHCPSubnet (self, context, sname, subnet_tuples):
    if not isinstance(subnet_tuples, list): subnet_tuples = [subnet_tuples]
    clid_map = {}
    for clid,subnet in subnet_tuples:
      clid_map[clid] = subnet

    return self._apiv3.poa(context, self.urlv3, sname, "api:uh.dhcp:set-subnet",
                         options = {"client-id-map" : clid_map})



  def addDNSResourceRecord (self, context, sname, client_id, record_name, record_type, record_value, record_ttl=7200):
    return self._apiv3.poa(context, self.urlv3, sname, "vts:uh.dnsroot:add-resource-record",
                           options = {"client-id" : client_id,
                           "record-name" : record_name,
                           "record-type" : record_type,
                           "record-value" : record_value,
                           "record-ttl" : record_ttl})

  def deleteDNSResourceRecord (self, context, sname, client_id, record_name, record_type):
    return self._apiv3.poa(context, self.urlv3, sname, "vts:uh.dnsroot:delete-resource-record",
                           options = {"client-id" : client_id,
                           "record-name" : record_name,
                           "record-type" : record_type})

  def getAllDNSResourceRecords(self, context, sname, client_ids):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self._apiv3.poa(context, self.urlv3, sname, "vts:uh.dnsroot:get-all-records",
                           options={"client-ids": client_ids})

  def getLastDNSDHCPops(self, context, sname, client_ids, number_of_operations, dns_OR_dhcp):
    if not isinstance(client_ids, list): client_ids = [client_ids]
    return self._apiv3.poa(context, self.urlv3, sname, "vts:uh.dnsdhcp:get-last-DNSDHCP-ops",
                           options={"client-ids": client_ids,
                           "number-of-operations": number_of_operations,
                           "dns-OR-dhcp": dns_OR_dhcp})


  def setDeleteLock (self, context, sname):
    """Prevent the given sliver from being deleted by another user with the credential.

    .. note::
      Locks are cumulative, and removed by calling `deletesliver`.  When the last locking
      user calls `deletesliver`, the sliver will be deleted.  It is not possible to remove
      your lock without risking deletion.

    Args:
      context: geni-lib context
      sname (str): Slice name
    """
    return self._apiv3.poa(context, self.urlv3, sname, "geni:set-delete-lock", {})

  def dropboxLink (self, context):
    """Link your user_urn to a Dropbox account at this aggregate.

    Args:
      context: geni-lib context

    Returns:
      str: Dropbox authorization URL to paste into web browser
    """
    return self._apiv3.paa(context, self.urlv3, "vts:dropbox:link-account")

  def dropboxFinalize (self, context, authcode):
    """Finalize the Dropbox account link for this aggregate.

    Args:
      context: geni-lib context
      authcode (str): Authorization code given by Dropbox
    """
    return self._apiv3.paa(context, self.urlv3, "vts:dropbox:complete-link", {"auth-code" : authcode})

  def dropboxUpload (self, context, sname, cvols):
    """Trigger upload to associated Dropbox account from requested container volumes.

    Args:
      context: geni-lib context
      sname (str): Slice name
      cvols (list): List of `(container client-id, volume-id)` tuples
    """
    data = {}
    for (cid,volid) in cvols:
      data.setdefault(cid, []).append(volid)
    return self._apiv3.poa(context, self.urlv3, sname, "vts:dropbox:upload", options = {"vols" : [data]})

  def hgPull (self, context, sname, cvols):
    """Update an HgMount volume with the latest data from the source repository.

    Args:
      context: geni-lib context
      sname (str): Slice name
      cvols (list): List of `(container client-id, volume-id)` tuples
    """

    data = {}
    for (cid,volid) in cvols:
      data.setdefault(cid, []).append((volid, True))
    return self._apiv3.poa(context, self.urlv3, sname, "vts:hg:pull", options = {"vols" : [data]})


Clemson = VTS("vts-clemson", "clemson.vts.bsswks.net")
GPO = VTS("vts-gpo", "gpo.vts.bsswks.net")
Illinois = VTS("vts-illinois", "uiuc.vts.bsswks.net")
NPS = VTS("vts-nps", "nps.vts.bsswks.net")
UKYPKS2 = VTS("vts-ukypks2", "ukypks2.vts.bsswks.net")
StarLight = VTS("vts-starlight", "starlight.vts.bsswks.net")


def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      yield obj

def name_to_aggregate ():
  result = dict()
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      result[obj.name] = obj
  return result

def aggregateFromHost (host):
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      if obj._host == host:
        return obj



# Copyright (c) 2014-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from lxml import etree as ET

import geni.namespaces as GNS
from geni.model.util import XPathXRange


VTSNS = GNS.Namespace("vts", "http://geni.bssoftworks.com/rspec/ext/vts/ad/1")
_XPNS = {'g' : GNS.REQUEST.name, 'v' : VTSNS.name}

def dumbcoerce (val):
  try:
    return int(val)
  except Exception:
    pass

  if val.lower() == "true": return True
  if val.lower() == "false": return False

  return val


class CircuitPlane(object):
  def __init__ (self):
    self.label = None
    self.type = None
    self.endpoint = None
    self.tunnel_types = []
    self.constraints = {}

  @classmethod
  def _fromdom (cls, elem):
    cp = CircuitPlane()
    cp.label = elem.get("label")
    supported = elem.xpath('v:supported-tunnels/v:tunnel-type', namespaces = _XPNS)

    for tuntyp in supported:
      cp.tunnel_types.append(tuntyp.get("name"))

    try:
      cp.endpoint = elem.xpath('v:endpoint', namespaces = _XPNS)[0].get("value")
    except IndexError:
      cp.endpoint = None

    for celem in elem.xpath('v:constraints/v:constraint', namespaces = _XPNS):
      cp.constraints[celem.get("key")] = dumbcoerce(celem.get("value"))

    return cp


class Image(object):
  def __init__ (self):
    self.name = None
    self.type = None
    self.mem_default = None
    self.mem_min = None
    self.mem_max = None

  def __str__ (self):
    return "[%s: %s]" % (self.type, self.name)

  @classmethod
  def _fromdom (cls, elem):
    img = cls()
    img.name = elem.get("name")
    img.type = elem.get("type")
    img.mem_default = int(elem.get("mem", 0))
    img.mem_min = int(elem.get("mem-min", img.mem_default))
    img.mem_max = int(elem.get("mem-max", img.mem_default))
    return img


class Advertisement(object):
  def __init__ (self, path = None, xml = None):
    if path:
      self._root = ET.parse(open(path))
    elif xml:
      self._root = ET.fromstring(xml)

  @property
  def circuit_planes (self):
    return XPathXRange(self._root.xpath("v:circuit-planes/v:circuit-plane", namespaces = _XPNS), CircuitPlane)

  @property
  def images (self):
    return XPathXRange(self._root.xpath("v:images/v:image", namespaces = _XPNS), Image)

  @property
  def text (self):
    return ET.tostring(self._root, pretty_print=True)



# Copyright (c) 2014  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


from __future__ import absolute_import

from lxml import etree as ET

from .. import namespaces as GNS
from .pg import Node

class DiskImage(object):
  def __init__ (self, name, version):
    self.name = name
    self.version = version

  def _write (self, element):
    di = ET.SubElement(element, "{%s}disk_image" % (GNS.REQUEST.name))
    di.attrib["name"] = self.name
    di.attrib["version"] = self.version
    return di

class XOSmall(Node):
  def __init__ (self, name, component_id = None):
    super(XOSmall, self).__init__(name, "xo.small", component_id, None)

class XOMedium(Node):
  def __init__ (self, name, component_id = None):
    super(XOMedium, self).__init__(name, "xo.medium", component_id, None)

class XOLarge(Node):
  def __init__ (self, name, component_id = None):
    super(XOLarge, self).__init__(name, "xo.large", component_id, None)



# Copyright (c) 2014-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os.path

#from cryptography import x509
#from cryptography.hazmat.backends import default_backend

from .core import FrameworkRegistry
from .. import tempfile

class KeyDecryptionError(Exception): pass

class ClearinghouseError(Exception):
  def __init__ (self, text, data = None):
    super(ClearinghouseError, self).__init__()
    self.text = text
    self.data = data
  def __str__ (self):
    return self.text


class Project(object):
  def __init__ (self, urn = None, uid = None, expired = None, role = None):
    self.expired = expired
    self.urn = urn
    self.uid = uid
    self.role = role

  def __str__ (self):
    if self.expired:
      return "[%s, %s, %s, EXPIRED]" % (self.urn, self.uid, self.role)
    return "[%s, %s, %s]" % (self.urn, self.uid, self.role)

  def __repr__ (self):
    return "<%s, %s>" % (self.urn, self.role)


class CHAPI2Project(Project):
  def __init__ (self, pinfo):
    if not pinfo.has_key("PROJECT_UID"):
      pinfo["PROJECT_UID"] = None

    if pinfo.has_key("EXPIRED"):
      super(CHAPI2Project, self).__init__(pinfo["PROJECT_URN"], pinfo["PROJECT_UID"],
                                          pinfo["EXPIRED"], pinfo["PROJECT_ROLE"])
    else:
      super(CHAPI2Project, self).__init__(pinfo["PROJECT_URN"], pinfo["PROJECT_UID"],
                                          pinfo["PROJECT_EXPIRED"])


class Member(object):
  def __init__ (self):
    self.urn = None
    self.uid = None
    self.email = None
    self.username = None
    self.firstname = None
    self.lastname = None
    self.emulab_role = None
    self.roles = {}

  @property
  def shortname (self):
    return self.urn.split("+")[-1]

  def _set_from_project (self, project_info):
    self.urn = project_info["PROJECT_MEMBER"]
    self.roles[project_info["PROJECT_URN"]] = project_info["PROJECT_ROLE"]

    try:
      self.uid = project_info["PROJECT_MEMBER_UID"]
    except KeyError:
      pass

    try:
      self.emulab_role = project_info["PROJECT_EMULAB_ROLE"]
    except KeyError:
      pass

  def _set_from_member (self, member_info):
    self.urn = member_info["MEMBER_URN"]
    self.uid = member_info["MEMBER_UID"]
    self.email = member_info["MEMBER_EMAIL"]
    self.username = member_info["MEMBER_USERNAME"]
    self.firstname = member_info["MEMBER_FIRSTNAME"]
    self.lastname = member_info["MEMBER_LASTNAME"]


class _MemberRegistry(object):
  def __init__ (self):
    self._members = {}

  def addProjectInfo (self, project_info):
    try:
      m = self._members[project_info["PROJECT_MEMBER"]]
    except KeyError:
      m = Member()
      self._members[project_info["PROJECT_MEMBER"]] = m

    m._set_from_project(project_info)
    return m

  def addMemberInfo (self, member_info):
    try:
      m = self._members[member_info["MEMBER_URN"]]
    except KeyError:
      m = Member()
      self._members[member_info["MEMBER_URN"]] = m

    m._set_from_member(member_info)
    return m


MemberRegistry = _MemberRegistry()

class Framework(object):
  class KeyPathError(Exception):
    def __init__ (self, path):
      super(Framework.KeyPathError, self).__init__()
      self.path = path
    def __str__ (self):
      return "Path %s does not contain a key" % (self.path)

  class UnconfiguredCertificateError(Exception):
    def __str__ (self):
      return "Key and Certificate must be set before using a framework"

  def __init__ (self, name = None):
    self.name = name
    self._type = None
    self._authority = None
    self._ch = None
    self._ma = None
    self._sa = None
    self._cert = None
    self._key_path = None
    self._key = None
    self._project = None
    self._userurn = None
    self._root_bundle = False

  @property
  def project (self):
    return self._project

  @project.setter
  def project (self, val):
    self._project = val

  @property
  def projecturn (self):
    # TODO:  Exception
    return None

  @property
  def key (self):
    if self._key is None:
      raise Framework.UnconfiguredCertificateError()
    return self._key

  @key.setter
  def key (self, path):
    self._key_path = path
    self._key = path

  def setKey (self, path, passwd):
    if not os.path.exists(path):
      raise Framework.KeyPathError(path)
    (tf, dpath) = tempfile.makeFile()

    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization

    try:
      key = serialization.load_pem_private_key(open(path, "rb").read(), passwd, default_backend())
    except ValueError:
      raise KeyDecryptionError()

    data = key.private_bytes(serialization.Encoding.PEM,
                             serialization.PrivateFormat.TraditionalOpenSSL,
                             serialization.NoEncryption())
    tf.write(data)
    tf.close()

    self._key = dpath

  @property
  def cert (self):
    if self._cert is None:
      raise Framework.UnconfiguredCertificateError()
    return self._cert

  @cert.setter
  def cert (self, val):
    self._cert = val

  @property
  def userurn (self):
    if not self._userurn:
      from cryptography import x509
      from cryptography.hazmat.backends import default_backend
      cert = x509.load_pem_x509_certificate(open(self._cert, "rb").read(), default_backend())
      for ext in cert.extensions:
        if ext.oid == x509.SubjectAlternativeName.oid:
          for uri in ext.value.get_values_for_type(x509.UniformResourceIdentifier):
            if uri.startswith("urn:publicid"):
              self._userurn = uri
              break
    return self._userurn


class ProtoGENI(Framework):
  SA = "https://www.emulab.net:12369/protogeni/xmlrpc/project/%s/sa"
  MA = "https://www.emulab.net:12369/protogeni/xmlrpc/project/%s/sa"

  def __init__ (self, name = "pg"):
    super(ProtoGENI, self).__init__("emulab")
    self._type = "pgch"
    self._ch = "https://www.emulab.net:12369/protogeni/xmlrpc/ch"
    self._sa = None
    self._ma = None

  @property
  def project (self):
    return super(ProtoGENI, self).project

  @project.setter
  def project (self, val):
    super(ProtoGENI, self.__class__).project.fset(self, val) # This is hinky
    self._sa = ProtoGENI.SA % (val)
    self._ma = ProtoGENI.MA % (val)

  def loadComponents (self, context):
    from ..minigcf import pgch1

    res = pgch1.ListComponents(self._ch, False, self.cert, self.key, context.ucred_pg)
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def getUserCredentials (self, owner_urn): # pylint: disable=unused-argument
    from ..minigcf import pgch1
    res = pgch1.GetCredential(self._sa, False, self.cert, self.key)
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)


class CHAPI1(Framework):
  def __init__ (self, name = "chapi"):
    super(CHAPI1, self).__init__(name)
    self._type = "chapi"


class CHAPI2(Framework):
  def __init__ (self, name = "chapi"):
    super(CHAPI2, self).__init__(name)
    self._type = "chapi"

  def projectNameToURN (self, name):
    ### TODO: Exception
    return None

  def sliceNameToURN (self, slice_name, project = None):
    ### TODO: Exception
    return None

  def loadAggregates (self):
    from ..minigcf import chapi2

    res = chapi2.lookup_service_info(self._ch, False, self.cert, self.key, [], "AGGREGATE_MANAGER")
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def createProject (self, context, name, exp, desc):
    from ..minigcf import chapi2

    res = chapi2.create_project(self._sa, False, self.cert, self.key, [context.ucred_api3], name, exp, desc)
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def listProjectMembers (self, context, project_urn = None):
    if not project_urn:
      project_urn = self.projecturn

    from ..minigcf import chapi2
    res = chapi2.lookup_project_members(self._sa, False, self.cert, self.key, [context.ucred_api3], project_urn)
    if res["code"] == 0:
      members = []
      for mobj in res["value"]:
        mobj["PROJECT_URN"] = project_urn
        members.append(MemberRegistry.addProjectInfo(mobj))
      return members
    else:
      raise ClearinghouseError(res["output"], res)

  def addProjectMembers (self, context, members, role = None, project = None):
    from ..minigcf import chapi2

    if not role: role = chapi2.PROJECT_ROLE.MEMBER
    if not project: project = context.project
    project_urn = self.projectNameToURN(project)

    res = chapi2.modify_project_membership(self._sa, False, self.cert, self.key, [context.ucred_api3],
                                           project_urn, add = [(x.urn, role) for x in members])
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def removeProjectMembers (self, context, members, project = None):
    from ..minigcf import chapi2

    if not project: project = context.project
    project_urn = self.projectNameToURN(project)

    res = chapi2.modify_project_membership(self._sa, False, self.cert, self.key, [context.ucred_api3],
                                           project_urn, remove = [x.urn for x in members])
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def listProjects (self, context, own = True, expired = False):
    from ..minigcf import chapi2

    if not own:
      res = chapi2.lookup_projects(self._sa, False, self.cert, self.key, [context.ucred_api3],
                                   expired = expired)
    else:
      res = chapi2.lookup_projects_for_member(self._sa, False, self.cert, self.key, [context.ucred_api3],
                                              context.userurn, expired = expired)

    if res["code"] == 0:
      projects = []
      if isinstance(res["value"], dict):
        for info in res["value"].values():
          projects.append(CHAPI2Project(info))
      else:
        for info in res["value"]:
          projects.append(CHAPI2Project(info))
      return projects
    else:
      raise ClearinghouseError(res["output"], res)

  def listAggregates (self, context):
    from ..minigcf import chapi2

    res = chapi2.lookup_aggregates(self._ch, False, self.cert, self.key)

    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def listSlices (self, context):
    from ..minigcf import chapi2
    res = chapi2.lookup_slices_for_project(self._sa, False, self.cert, self.key,
                                           [context.ucred_api3], context.project_urn)
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def listSliceMembers (self, context, slicename):
    from ..minigcf import chapi2

    slice_urn = self.sliceNameToURN(slicename)

    res = chapi2.lookup_slice_members(self._sa, False, self.cert, self.key,
                                      [context.ucred_api3], slice_urn)
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def addSliceMembers (self, context, slicename, members, role = None):
    from ..minigcf import chapi2

    if not role:
      role = chapi2.SLICE_ROLE.MEMBER

    if not isinstance(members, (list,set,tuple)):
      members = [members]

    slice_urn = self.sliceNameToURN(slicename)

    res = chapi2.modify_slice_membership(self._sa, False, self.cert, self.key, [context.ucred_api3],
                                         slice_urn, add = [(x.urn, role) for x in members])

    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def removeSliceMembers (self, context, slicename, members):
    from ..minigcf import chapi2

    if not isinstance(members, (list,set,tuple)):
      members = [members]

    slice_urn = self.sliceNameToURN(slicename)

    res = chapi2.modify_slice_membership(self._sa, False, self.cert, self.key, [context.ucred_api3],
                                         slice_urn, remove = [x.urn for x in members])

    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def getUserCredentials (self, owner_urn):
    from ..minigcf import chapi2
    res = chapi2.get_credentials(self._ma, False, self.cert, self.key, [], owner_urn)
    if res["code"] == 0:
      return res["value"][0]["geni_value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def getSliceCredentials (self, context, slicename):
    from ..minigcf import chapi2

    slice_urn = self.sliceNameToURN(slicename)

    res = chapi2.get_credentials(self._sa, False, self.cert, self.key, [context.ucred_api3], slice_urn)
    if res["code"] == 0:
      return res["value"][0]["geni_value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def createSlice (self, context, slicename, project_urn = None, exp = None, desc = None):
    from ..minigcf import chapi2

    if project_urn is None:
      project_urn = self.projectNameToURN(context.project)

    res = chapi2.create_slice(self._sa, False, self.cert, self.key, [context.ucred_api3], slicename, project_urn, exp, desc)
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def renewSlice (self, context, slicename, exp):
    from ..minigcf import chapi2

    fields = {"SLICE_EXPIRATION" : exp.strftime(chapi2.DATE_FMT)}
    slice_urn = self.sliceNameToURN(slicename)
    slice_info = context.getSliceInfo(slicename)

    res = chapi2.update_slice(self._sa, False, self.cert, self.key,
                              [slice_info.cred_api3, context.ucred_api3],
                              slice_urn, fields)
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def lookupSSHKeys (self, context, user_urn):
    from ..minigcf import chapi2

    res = chapi2.lookup_key_info(self._ma, False, self.cert, self.key, [context.ucred_api3], user_urn)
    if res["code"] == 0:
      key_list = [x["KEY_PUBLIC"] for x in res["value"].values()]
      return key_list
    else:
      raise ClearinghouseError(res["output"], res)

  def lookupMemberInfo (self, context, urn = None, uid = None):
    from ..minigcf import chapi2

    res = chapi2.lookup_member_info(self._ma, False, self.cert, self.key, [context.ucred_api3],
                                    urn = urn, uid = uid)
    return MemberRegistry.addMemberInfo(res["value"].values()[0])


class Portal(CHAPI2):
  def __init__ (self):
    super(Portal, self).__init__("gpo-ch2")
    self._authority = "ch.geni.net"
    self._ch = "https://ch.geni.net:8444/CH"
    self._ma = "https://ch.geni.net:443/MA"
    self._sa = "https://ch.geni.net:443/SA"
    self._memberuid = None
    self._project_info = {}

  @property
  def projecturn (self):
    return self.projectNameToURN(self.project)

  def projectInfo (self, context):
    purn = self.projecturn
    if not self._project_info.has_key(purn):
      from ..minigcf import chapi2
      projects = chapi2.lookup_projects(self._sa, self._root_bundle, self.cert, self.key,
                                        [context.ucred_api3], purn)
      self._project_info[purn] = CHAPI2Project(projects["value"][purn])
    return self._project_info[purn]

  def projectNameToURN (self, name):
    return "urn:publicid:IDN+ch.geni.net+project+%s" % (name)

  def sliceNameToURN (self, name, project = None):
    if not project:
      project = self.project
    return "urn:publicid:IDN+ch.geni.net:%s+slice+%s" % (project, name)

  def _getMemberUID (self, context):
    from ..minigcf import chapi2
    if not self._memberuid:
      infodict = chapi2.lookup_member_info(self._ma, self._root_bundle, self.cert, self.key,
                                           [context.ucred_api3], self.userurn)
      minfo = infodict["value"][self.userurn]
      self._memberuid = minfo["MEMBER_UID"]
    return self._memberuid

  def getPendingProjectRequests (self, context):
    from ..minigcf import chapi2
    res = chapi2.get_pending_requests(self._sa, self._root_bundle, self.cert, self.key,
                                      [context.ucred_api3], self._getMemberUID(context),
                                      self.projectInfo(context).uid)
    if res["code"] == 0:
      return res["value"]
    else:
      raise ClearinghouseError(res["output"], res)

  def resolveRequests (self, context, req_ids, status, desc):
    from ..minigcf import chapi2
    for rid in req_ids:
      res = chapi2.resolve_request(self._sa, self._root_bundle, self.cert, self.key,
                                   [context.ucred_api3], rid, status, desc)
      if res["code"] != 0:
        raise ClearinghouseError(res["output"], res)
    return


class EmulabCH2(CHAPI2):
  SA = "https://www.emulab.net:12369/protogeni/xmlrpc/geni-sa/2"
#  SA = "https://www.emulab.net:12369/protogeni/xmlrpc/project/%s/geni-sa/2"
  MA = "https://www.emulab.net:12369/protogeni/xmlrpc/geni-ma"
#  MA = "https://www.emulab.net:12369/protogeni/xmlrpc/project/%s/geni-ma"

  def __init__ (self):
    super(EmulabCH2, self).__init__("emulab-ch2")
    self._authority = ""
    self._ch = EmulabCH2.SA
    self._sa = EmulabCH2.SA
    self._ma = EmulabCH2.MA

  @property
  def projecturn (self):
    return self.projectNameToURN(self.project)

  def projectNameToURN (self, name):
    return "urn:publicid:IDN+emulab.net+project+%s" % (name)

  def sliceNameToURN (self, name, project = None):
    if not project:
      project = self.project
    return "urn:publicid:IDN+emulab.net:%s+slice+%s" % (project, name)

  @property
  def project (self):
    return super(EmulabCH2, self).project

  @project.setter
  def project (self, val):
    self._project = val
    self._sa = EmulabCH2.SA
    self._ma = EmulabCH2.MA
#    self._ma = EmulabCH2.MA % (val)

#  def listSlices (self, context):
#    from ..minigcf import chapi2
#    res = chapi2._lookup(self._sa, False, self.cert, self.key, "SLICE", [context.ucred_api3], {})
#    if res["code"] == 0:
#      return res["value"]
#    else:
#      raise ClearinghouseError(res["output"], res)

FrameworkRegistry.register("portal", Portal)
FrameworkRegistry.register("gpo-ch2", Portal)
FrameworkRegistry.register("pg", ProtoGENI)
FrameworkRegistry.register("emulab-ch2", EmulabCH2)



# Copyright (c) 2017  Barnstormer Softworks, Ltd.

#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import xmlrpclib

import requests

from .. import _coreutil as GCU
from . import config

GCU.disableUrllibWarnings()

def headers ():
  return GCU.defaultHeaders()

# pylint: disable=unsubscriptable-object
def _rpcpost (url, req_data, cert, root_bundle):
  if isinstance(config.HTTP.LOG_URLS, tuple):
    config.HTTP.LOG_URLS[0].log(config.HTTP.LOG_URLS[1], "POST: %s" % (url))
  s = requests.Session()
  s.mount(url, GCU.TLSHttpAdapter())
  if isinstance(config.HTTP.LOG_RAW_REQUESTS, tuple):
    config.HTTP.LOG_RAW_REQUESTS[0].log(config.HTTP.LOG_RAW_REQUESTS[1], req_data)
  resp = s.post(url, req_data, cert=cert, verify=root_bundle, headers = headers(),
                timeout = config.HTTP.TIMEOUT, allow_redirects = config.HTTP.ALLOW_REDIRECTS)
  if resp.status_code != 200:
    resp.raise_for_status()
  if isinstance(config.HTTP.LOG_RAW_RESPONSES, tuple):
    config.HTTP.LOG_RAW_RESPONSES[0].log(config.HTTP.LOG_RAW_RESPONSES[1], resp.content)
  return xmlrpclib.loads(resp.content, use_datetime=True)[0][0]
  



# Copyright (c) 2013-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os

from lxml import etree as ET

from .pg import Link
from .. import namespaces as GNS
from .pg import Namespaces as PGNS
from ..model.util import XPathXRange

_XPNS = {'g' : GNS.REQUEST.name, 's' : GNS.SVLAN.name, 'e' : PGNS.EMULAB.name,
         'i' : PGNS.INFO.name, 'p' : PGNS.PARAMS.name, 'u' : GNS.USER.name}

class ManifestLink(Link):
  def __init__ (self):
    super(ManifestLink, self).__init__()
    self.interface_refs = []
    self.sliver_id = None
    self.vlan = None
    self._elem = None

  @classmethod
  def _fromdom (cls, elem):
    lnk = ManifestLink()
    lnk._elem = elem
    lnk.client_id = elem.get("client_id")
    lnk.sliver_id = elem.get("sliver_id")
    lnk.vlan = elem.get("vlantag", None)

    refs = elem.xpath('g:interface_ref', namespaces = _XPNS)
    for ref in refs:
      lnk.interface_refs.append(ref.get("sliver_id"))

    svlans = elem.xpath('s:link_shared_vlan', namespaces = _XPNS)
    if svlans:
      # TODO: Can a link be attached to more than one shared vlan?
      # Don't believe PG supports trunks, but the rspec doesn't really forbid it
      lnk.vlan = svlans[0].get("name")

    return lnk

  @property
  def text (self):
    return ET.tostring(self._elem, pretty_print=True)


class ManifestSvcLogin(object):
  def __init__ (self):
    self.auth = None
    self.hostname = None
    self.port = None
    self.username = None

  @classmethod
  def _fromdom (cls, elem):
    n = ManifestSvcLogin()
    n.auth = elem.get("authentication")
    n.hostname = elem.get("hostname")
    n.port = int(elem.get("port"))
    n.username = elem.get("username")

    return n


class ManifestSvcUser(object):
  def __init__ (self):
    self.login = None
    self.public_key = None

  @classmethod
  def _fromdom (cls, elem):
    n = cls()
    n.login = elem.get("login")
    pkelems = elem.xpath('u:public_key', namespaces = _XPNS)
    if pkelems:
      n.public_key = pkelems[0].text.strip()
    return n


class ManifestNode(object):
  class Interface(object):
    def __init__ (self):
      self.client_id = None
      self.mac_address = None
      self.sliver_id = None
      self.address_info = None
      self.component_id = None

  def __init__ (self):
    super(ManifestNode, self).__init__()
    self.logins = []
    self.users = []
    self.interfaces = []
    self.client_id = None
    self.component_id = None
    self.sliver_id = None
    self._elem = None
    self._hostfqdn = None
    self._hostipv4 = None

  @property
  def name (self):
    return self.client_id

  @property
  def hostfqdn (self):
    if not self._hostfqdn:
      self._populateHostInfo()
    return self._hostfqdn

  @property
  def hostipv4 (self):
    if not self._hostipv4:
      self._populateHostInfo()
    return self._hostipv4

  def _populateHostInfo (self):
    host = self._elem.xpath('g:host', namespaces = _XPNS)
    if host:
      self._hostfqdn = host[0].get("name", None)
      self._hostipv4 = host[0].get("ipv4", None)

  @classmethod
  def _fromdom (cls, elem):
    n = ManifestNode()
    n._elem = elem
    n.client_id = elem.get("client_id")
    n.component_id = elem.get("component_id")
    n.sliver_id = elem.get("sliver_id")

    logins = elem.xpath('g:services/g:login', namespaces = _XPNS)
    for lelem in logins:
      l = ManifestSvcLogin._fromdom(lelem)
      n.logins.append(l)

    users = elem.xpath('g:services/u:services_user', namespaces = _XPNS)
    for uelem in users:
      u = ManifestSvcUser._fromdom(uelem)
      n.users.append(u)

    interfaces = elem.xpath('g:interface', namespaces = _XPNS)
    for ielem in interfaces:
      i = ManifestNode.Interface()
      i.client_id = ielem.get("client_id")
      i.sliver_id = ielem.get("sliver_id")
      i.component_id = ielem.get("component_id")
      i.mac_address = ielem.get("mac_address")
      try:
        ipelem = ielem.xpath('g:ip', namespaces = _XPNS)[0]
        i.address_info = (ipelem.get("address"), ipelem.get("netmask"))
      except Exception:
        pass
      n.interfaces.append(i)

    return n

  @property
  def text (self):
    return ET.tostring(self._elem, pretty_print=True)


class ManifestParameter(object):
  def __init__ (self, name, value):
    super(ManifestParameter, self).__init__()
    self.name = name
    self.value = value

  @classmethod
  def _fromdom (cls, elem):
    n = ManifestParameter(elem.get('name'), elem.get('value'))
    return n


class Manifest(object):
  def __init__ (self, path = None, xml = None):
    if path:
      self._xml = open(path, "r").read()
    elif xml:
      self._xml = xml
    self._root = ET.fromstring(self._xml)
    self._pid = os.getpid()

  @property
  def root (self):
    if os.getpid() != self._pid:
      self._root = ET.fromstring(self._xml)
      self._pid = os.getpid()
    return self._root

  @property
  def latitude (self):
    loc = self._root.xpath('i:site_info/i:location', namespaces = _XPNS)
    if loc:
      return loc[0].get("latitude")

  @property
  def longitude (self):
    loc = self._root.xpath('i:site_info/i:location', namespaces = _XPNS)
    if loc:
      return loc[0].get("longitude")

  @property
  def expiresstr (self):
    return self._root.get("expires")

  @property
  def links (self):
    return XPathXRange(self.root.findall("{%s}link" % (GNS.REQUEST)), ManifestLink)

  @property
  def nodes (self):
    return XPathXRange(self.root.findall("{%s}node" % (GNS.REQUEST)), ManifestNode)

  @property
  def parameters (self):
    for param in self.root.findall("{%s}data_set/{%s}data_item"
                                   % (PGNS.PARAMS.name,PGNS.PARAMS.name)):
      yield ManifestParameter._fromdom(param)

  @property
  def text (self):
    return ET.tostring(self.root, pretty_print=True)

  def _repr_html_ (self):
    return """
<table>
  <tr><th scope="row">Expires</th><td>%s</td></tr>
  <tr><th scope="row">Nodes</th><td>%d</td></tr>
  <tr><th scope="row">Links</th><td>%d</td></tr>
  <tr><th scope="row">Location</th><td>%s, %s</td></tr>
</table>""" % (self.expiresstr, len(self.nodes), len(self.links),
               self.latitude, self.longitude)

  def write (self, path):
    """
.. deprecated:: 0.4
    Use :py:meth:`geni.rspec.pg.Request.writeXML` instead."""

    import geni.warnings as GW
    import warnings
    warnings.warn("The Manifest.write() method is deprecated, please use Manifest.writeXML() instead",
                  GW.GENILibDeprecationWarning, 2)
    self.writeXML(path)

  def writeXML (self, path):
    """Write the current manifest as an XML file that contains an rspec in the format returned by the
    aggregate."""
    f = open(path, "w+")
    f.write(ET.tostring(self.root, pretty_print=True))
    f.close()



# Copyright (c) 2014-2016  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json

import requests
import requests.auth

import requests.packages.urllib3
import requests.packages.urllib3.exceptions
requests.packages.urllib3.disable_warnings((requests.packages.urllib3.exceptions.InsecureRequestWarning,
                                            requests.packages.urllib3.exceptions.InsecurePlatformWarning))

class Connection(object):
  def __init__ (self):
    self.user = "germadmin"
    self.password = None
    self.host = "localhost"
    self.port = 3626

  @property
  def baseurl (self):
    return "https://%s:%d" % (self.host, self.port)

  @property
  def rkwargs (self):
    d = {"headers" : self.headers,
         "auth" : self.auth,
         "verify" : False}
    return d

  @property
  def headers (self):
    return {"Content-Type" : "application/json"}

  @property
  def auth (self):
    return requests.auth.HTTPBasicAuth(self.user, self.password)

  @property
  def configkeys (self):
    url = "%s/core/admin/get-config-keys" % (self.baseurl)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  @property
  def version (self):
    url = "%s/core/admin/get-version" % (self.baseurl)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]["version"]

  def getConfig (self, key):
    url = "%s/core/admin/get-config" % (self.baseurl)
    d = json.dumps({"key":key})
    r = requests.post(url, d, **self.rkwargs)
    return r.json()["value"]

  def setConfig (self, key, value):
    url = "%s/core/admin/set-config" % (self.baseurl)
    d = json.dumps({"key":key, "value":value})
    r = requests.post(url, d, **self.rkwargs)
    return r.json()["value"]

  def setLocation (self, country, latitude, longitude):
    url = "%s/core/admin/set-location" % (self.baseurl)
    d = json.dumps({"country" : country, "long" : longitude, "lat" : latitude})
    r = requests.post(url, d, **self.rkwargs)
    return r.json()["value"]
